import logging
from dataclasses import dataclass
from typing import cast, List, Tuple, Dict, Any, Optional, Union, NamedTuple

import hydra
import numpy as np
import pandas as pd
import torch
from hydra.utils import instantiate
from omegaconf import DictConfig
from pydantic import BaseModel
from pytorch_lightning import Trainer, LightningModule
from pytorch_lightning.utilities.types import (
    STEP_OUTPUT,
    TRAIN_DATALOADERS,
    EVAL_DATALOADERS,
    EPOCH_OUTPUT,
)
from torch import optim
from torch.utils.data import DataLoader, Dataset
from transformers import (
    AutoConfig,
    AutoTokenizer,
    AutoModel,
    BatchEncoding,
    PreTrainedTokenizerBase,
    DataCollatorWithPadding,
)
from transformers.file_utils import PaddingStrategy
from tokenizers import Encoding

from kazu.utils.constants import HYDRA_VERSION_BASE

logger = logging.getLogger(__name__)


@dataclass
class SapbertDataCollatorWithPadding:
    """
    data collator to be used with HFSapbertPairwiseDataset
    """

    tokenizer: PreTrainedTokenizerBase
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None

    def __call__(self, features: List) -> Tuple[BatchEncoding, BatchEncoding]:
        query_toks1 = [x["query_toks1"] for x in features]
        query_toks1_enc = self.tokenizer.pad(
            query_toks1,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )
        query_toks2 = [x["query_toks2"] for x in features]
        query_toks2_enc = self.tokenizer.pad(
            query_toks2,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        return query_toks1_enc, query_toks2_enc


def init_hf_collate_fn(tokenizer: PreTrainedTokenizerBase) -> DataCollatorWithPadding:
    """
    get a standard HF DataCollatorWithPadding, with padding=PaddingStrategy.LONGEST

    :param tokenizer:
    :return:
    """
    collate_func = DataCollatorWithPadding(tokenizer=tokenizer, padding=PaddingStrategy.LONGEST)
    return collate_func


class HFSapbertInferenceDataset(Dataset):
    """
    A dataset to be used for inferencing. In addition to standard BERT  encodings, this
    uses an 'indices' encoding that can be used to track the vector index of an embedding.
    This is needed in a multi GPU environment
    """

    def __getitem__(self, index) -> Dict[str, Any]:
        query_toks1 = {
            "input_ids": self.encodings.data["input_ids"][index],
            "token_type_ids": self.encodings.data["token_type_ids"][index],
            "attention_mask": self.encodings.data["attention_mask"][index],
            "indices": self.encodings.data["indices"][index],
        }
        return query_toks1

    def __init__(self, encodings: BatchEncoding):
        """
        simple implementation of IterableDataset, producing HF tokenizer input_id

        :param encodings:
        """
        self.encodings = encodings

    def __len__(self):
        encodings = cast(List[Encoding], self.encodings.encodings)
        return len(encodings)


class HFSapbertPairwiseDataset(Dataset):
    """
    Dataset used for training SapBert.
    """

    def __getitem__(self, index) -> Dict[str, Any]:
        query_toks1 = {
            "input_ids": self.encodings_1.data["input_ids"][index],
            "token_type_ids": self.encodings_1.data["token_type_ids"][index],
            "labels": self.encodings_1.data["labels"][index],
            "attention_mask": self.encodings_1.data["attention_mask"][index],
        }
        query_toks2 = {
            "input_ids": self.encodings_2.data["input_ids"][index],
            "token_type_ids": self.encodings_2.data["token_type_ids"][index],
            "labels": self.encodings_2.data["labels"][index],
            "attention_mask": self.encodings_2.data["attention_mask"][index],
        }

        return {"query_toks1": query_toks1, "query_toks2": query_toks2}

    def __init__(self, encodings_1: BatchEncoding, encodings_2: BatchEncoding, labels: np.ndarray):
        """
        :param encodings_1: encodings for example 1
        :param encodings_2: encodings for example 2
        :param labels: labels i.e. knowledgebase identifier for both encodings, as an int
        """
        encodings_1["labels"] = labels
        encodings_2["labels"] = labels
        self.encodings_1 = encodings_1
        self.encodings_2 = encodings_2

    def __len__(self):
        encodings = cast(List[Encoding], self.encodings_1.encodings)
        return len(encodings)


class SapbertTrainingParams(BaseModel):
    topk: int  # score this many nearest neighbours against target
    lr: float  # learning rate
    weight_decay: float
    miner_margin: float  # passed to TripletMarginMiner
    type_of_triplets: str  # passed to TripletMarginMiner
    train_file: str  # a parquet file with three columns - syn1, syn2 and id
    train_batch_size: int
    num_workers: int  # passed to dataloaders


def get_embedding_dataloader_from_strings(
    texts: List[str],
    tokenizer: PreTrainedTokenizerBase,
    batch_size: int,
    num_workers: int,
    max_length: int = 50,
):
    """
    get a dataloader with dataset HFSapbertInferenceDataset and DataCollatorWithPadding. This should be used to
    generate embeddings for strings of interest

    :param texts: strings to use in the dataset
    :param tokenizer:
    :param batch_size:
    :param num_workers:
    :return:
    """
    indices = [i for i in range(len(texts))]
    # padding handled by collate func
    batch_encodings = tokenizer(
        texts, padding=PaddingStrategy.MAX_LENGTH, max_length=max_length, truncation=True
    )
    batch_encodings["indices"] = indices
    dataset = HFSapbertInferenceDataset(batch_encodings)
    collate_func = init_hf_collate_fn(tokenizer)
    loader = DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        collate_fn=collate_func,
        num_workers=num_workers,
    )
    return loader


class SapbertEvaluationDataset(NamedTuple):
    """
    to evaluate a given embedding model, we need a query datasource (i.e. things that need to be linked)]
    and an ontology datasource (i.e. things that we need to generate an embedding space for, that can be queried
    against)
    each should have three columns: default_label (text), iri (ontology id) and source (ontology name)
    """

    query_source: pd.DataFrame
    ontology_source: pd.DataFrame


class Candidate(NamedTuple):
    """
    a knowledgebase entry
    """

    default_label: str
    iri: str
    correct: bool


class GoldStandardExample(NamedTuple):
    gold_default_label: str
    gold_iri: str
    candidates: List[
        Candidate
    ]  # candidates (aka nearest neighbours) associate with this gold instance


class SapbertEvaluationDataManager:
    """
    manages the loading/parsing of multiple evaluation datasets. Each dataset should have two sources, a query source
    and an ontology source. these are then converted into data loaders, while maintaining a reference to the
    embedding metadata that should be used for evaluation

    self.dataset is Dict[dataset_name,SapbertEvaluationDataset] after construction
    """

    def __init__(self, sources: Dict[str, List[str]], debug=False):
        self.datasets: Dict[str, SapbertEvaluationDataset] = {}
        for source_name, (
            query_source_path,
            ontology_source_path,
        ) in sources.items():
            query_source = pd.read_parquet(query_source_path)
            ontology_source = pd.read_parquet(ontology_source_path)
            if debug:
                query_source = query_source.head(10)
                ontology_source = ontology_source.sample(frac=1.0).head(100)
            self.datasets[source_name] = SapbertEvaluationDataset(
                query_source=query_source, ontology_source=ontology_source
            )


class PLSapbertModel(LightningModule):
    """
    Pytorch lightning production implementation of SapBert.

    Original source:

    https://github.com/cambridgeltl/sapbert

    Licensed under MIT

    Copyright (c) Facebook, Inc. and its affiliates.

    .. raw:: html

        <details>
        <summary>Full License</summary>

    MIT License

    Copyright (c) Facebook, Inc. and its affiliates.

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    .. raw:: html

        </details>

    Paper:

    Fangyu Liu, Ehsan Shareghi, Zaiqiao Meng, Marco Basaldella, and Nigel Collier. 2021.
    `Self-alignment pretraining for biomedical entity representations. <https://www.aclweb.org/anthology/2021.naacl-main.334.pdf>`_
    In Proceedings of the 2021 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies, pages 4228–4238.

    .. raw:: html

        <details>
        <summary>Bibtex Citation Details</summary>

    .. code:: bibtex

        @inproceedings{liu2021self,
            title={Self-Alignment Pretraining for Biomedical Entity Representations},
            author={Liu, Fangyu and Shareghi, Ehsan and Meng, Zaiqiao and Basaldella, Marco and Collier, Nigel},
            booktitle={Proceedings of the 2021 Conference of the North American Chapter of the Association for Computational Linguistics: Human Language Technologies},
            pages={4228--4238},
            month = jun,
            year={2021}
        }

    .. raw:: html

        </details>

    """

    def __init__(
        self,
        model_name_or_path: str,
        sapbert_training_params: Optional[SapbertTrainingParams] = None,
        sapbert_evaluation_manager: Optional[SapbertEvaluationDataManager] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        :param model_name_or_path: passed to AutoModel.from_pretrained
        :param sapbert_training_params: optional SapbertTrainingParams, only needed if training a model
        :param sapbert_evaluation_manager: optional SapbertEvaluationDataManager, only needed if training a model
        :param args: passed to LightningModule
        :param kwargs: passed to LightningModule
        """

        super().__init__(*args, **kwargs)

        self.config = AutoConfig.from_pretrained(model_name_or_path)
        self.tokeniser = AutoTokenizer.from_pretrained(model_name_or_path, config=self.config)
        self.model = AutoModel.from_pretrained(model_name_or_path, config=self.config)
        self.sapbert_evaluation_manager = sapbert_evaluation_manager
        self.sapbert_training_params = sapbert_training_params
        if sapbert_training_params is not None:
            try:
                from pytorch_metric_learning import miners, losses
            except ImportError as e:
                raise ImportError(
                    "Running the SapBERT model training code requires pytorch_metric_learning to be installed.\n"
                    "We recommend running 'pip install kazu[model_training]' to get all model training"
                    " dependencies."
                ) from e
            self.loss = losses.MultiSimilarityLoss(alpha=1, beta=60, base=0.5)
            self.miner = miners.TripletMarginMiner(
                margin=sapbert_training_params.miner_margin,
                type_of_triplets=sapbert_training_params.type_of_triplets,
            )
            self.ontology_embeddings = None

    def configure_optimizers(self):
        """Implementation of :external+pytorch_lightning:ref:`LightningModule.configure_optimizers </common/lightning_module.rst#configure-optimizers>`\\ ."""
        assert self.sapbert_training_params is not None
        optimizer = optim.AdamW(
            [
                {"params": self.model.parameters()},
            ],
            lr=self.sapbert_training_params.lr,
            weight_decay=self.sapbert_training_params.weight_decay,
        )
        return optimizer

    def forward(self, batch):
        """
        for inference

        :param batch: standard bert input, with an additional 'indices' for representing the location of the embedding
        :return:
        """
        indices = batch.pop("indices")
        batch_embeddings = self.model(**batch)
        batch_embeddings = batch_embeddings.last_hidden_state[:, 0, :]  # cls token
        # put index as dict key so we can realign the embedding space
        return {
            index.item(): batch_embeddings[[batch_index], :]
            for batch_index, index in enumerate(indices)
        }

    def training_step(self, batch, batch_idx, *args, **kwargs) -> STEP_OUTPUT:
        """Implementation of :external+pytorch_lightning:ref:`LightningModule.training_step </common/lightning_module.rst#training-step>`\\ ."""
        query_toks1, query_toks2 = batch
        # labels should be identical, so we only need one
        labels = query_toks1.pop("labels")
        query_toks2.pop("labels")
        last_hidden_state1 = self.model(**query_toks1, return_dict=True).last_hidden_state
        last_hidden_state2 = self.model(**query_toks2, return_dict=True).last_hidden_state
        query_embed1 = last_hidden_state1[:, 0]  # query : [batch_size, hidden]
        query_embed2 = last_hidden_state2[:, 0]  # query : [batch_size, hidden]
        query_embed = torch.cat([query_embed1, query_embed2], dim=0)
        labels = torch.cat([labels, labels], dim=0)
        hard_pairs = self.miner(query_embed, labels)
        return self.loss(query_embed, labels, hard_pairs)

    def train_dataloader(self) -> TRAIN_DATALOADERS:
        """Implementation of :external+pytorch_lightning:ref:`LightningModule.train_dataloader </common/lightning_module.rst#train-dataloader>`\\ ."""
        assert self.sapbert_training_params is not None
        training_df = pd.read_parquet(self.sapbert_training_params.train_file)
        labels = training_df["id"].astype("category").cat.codes.to_numpy()
        encodings_1 = self.tokeniser(training_df["syn1"].tolist())
        encodings_2 = self.tokeniser(training_df["syn2"].tolist())
        encodings_1["labels"] = labels
        encodings_2["labels"] = labels
        train_set = HFSapbertPairwiseDataset(
            labels=labels, encodings_1=encodings_1, encodings_2=encodings_2
        )
        train_loader = torch.utils.data.DataLoader(
            train_set,
            batch_size=self.sapbert_training_params.train_batch_size,
            shuffle=True,
            num_workers=self.sapbert_training_params.num_workers,
            collate_fn=SapbertDataCollatorWithPadding(
                self.tokeniser, padding=PaddingStrategy.LONGEST
            ),
        )

        return train_loader

    def val_dataloader(self) -> EVAL_DATALOADERS:
        """Implementation of :external+pytorch_lightning:ref:`LightningModule.val_dataloader </common/lightning_module.rst#val-dataloader>`\\ ."""
        dataloaders = []
        assert self.sapbert_evaluation_manager is not None
        assert self.sapbert_training_params is not None
        for query_source, ontology_source in self.sapbert_evaluation_manager.datasets.values():
            query_dataloader = get_embedding_dataloader_from_strings(
                texts=query_source["default_label"].tolist(),
                tokenizer=self.tokeniser,
                batch_size=self.sapbert_training_params.train_batch_size,
                num_workers=self.sapbert_training_params.num_workers,
            )
            ontology_dataloader = get_embedding_dataloader_from_strings(
                texts=ontology_source["default_label"].tolist(),
                tokenizer=self.tokeniser,
                batch_size=self.sapbert_training_params.train_batch_size,
                num_workers=self.sapbert_training_params.num_workers,
            )
            dataloaders.append(query_dataloader)
            dataloaders.append(ontology_dataloader)

        return dataloaders

    def validation_step(self, batch, batch_idx, dataset_idx) -> Optional[STEP_OUTPUT]:
        """Implementation of :external+pytorch_lightning:ref:`LightningModule.validation_step </common/lightning_module.rst#validation-step>`\\ ."""
        return self(batch)

    def predict_step(self, batch: Any, batch_idx: int, dataloader_idx: Optional[int] = None) -> Any:
        """Implementation of :external+pytorch_lightning:ref:`LightningModule.predict_step </common/lightning_module.rst#predict-step>`\\ ."""
        return self(batch)

    def get_embeddings(self, output: List[Dict[int, torch.Tensor]]) -> torch.Tensor:
        """
        get a tensor of embeddings in original order

        :param output: int is the original index of the input (i.e. what comes out of self.forward)
        :return:
        """
        full_dict = {}
        for batch in output:
            full_dict.update(batch)
        if len(full_dict) > 1:
            embedding = torch.squeeze(torch.cat(list(full_dict.values())))
        else:
            embedding = torch.cat(list(full_dict.values()))
        return embedding

    def validation_epoch_end(self, outputs: Union[EPOCH_OUTPUT, List[EPOCH_OUTPUT]]) -> None:
        """
        lightning override
        generate new embeddings for each :attr:`SapbertEvaluationDataset.ontology_source` and query them with
        :attr:`SapbertEvaluationDataset.query_source`

        :param outputs:
        :return:
        """
        assert self.sapbert_evaluation_manager is not None
        raise NotImplementedError

    def log_results(self, dataset_name, metrics):
        for key, val in metrics.items():
            if key.startswith("acc"):
                self.log(key, value=val, rank_zero_only=True)
                logger.info(f"{dataset_name}: {key}, {val}")

    def get_candidate_dict(self, np_candidates: pd.DataFrame, golden_iri: str) -> List[Candidate]:
        """
        Convert rows in a dataframe representing candidate KB entries into a corresponding
        :class:`Candidate` per row

        :param np_candidates:
        :param golden_iri:
        :return:
        """
        candidates_filtered = []
        for i, np_candidate_row in np_candidates.iterrows():
            candidates_filtered.append(
                Candidate(
                    default_label=np_candidate_row["default_label"],
                    iri=np_candidate_row["iri"],
                    correct=np_candidate_row["iri"] == golden_iri,
                )
            )
        return candidates_filtered

    def evaluate_topk_acc(self, queries: List[GoldStandardExample]) -> Dict[str, float]:
        """
        get a dictionary of accuracy results at different levels of k (nearest neighbours)

        :param queries:
        :return:
        """
        k = len(queries[0].candidates)
        result = {}
        for i in range(0, k):
            hit = 0
            for query in queries:
                candidates = query.candidates[: i + 1]  # to get acc@(i+1)
                if any(candidate.correct for candidate in candidates):
                    hit += 1
            result["acc{}".format(i + 1)] = hit / len(queries)

        return result

    def get_embeddings_for_strings(
        self, texts: List[str], trainer: Optional[Trainer] = None, batch_size: Optional[int] = None
    ) -> torch.Tensor:
        """
        For a list of strings, generate embeddings.

        This is a convenience function for users, as we need to carry out these steps several times
        in the codebase.

        :param texts:
        :param trainer: an optional PL Trainer to use. If not specified, uses the default one
        :param batch_size: optional batch size to use. If not specified, use 16
        :return: a 2d tensor of embeddings
        """
        if trainer is None:
            trainer = Trainer(enable_progress_bar=False, logger=False)
        if batch_size is None:
            batch_size = 16

        loader = get_embedding_dataloader_from_strings(texts, self.tokeniser, batch_size, 0)
        results = self.get_embeddings_from_dataloader(loader, trainer)
        return results

    def get_embeddings_from_dataloader(self, loader: DataLoader, trainer: Trainer) -> torch.Tensor:
        """
        get the cls token output from all data in a dataloader as a 2d tensor

        :param loader:
        :param trainer: the PL Trainer to use
        :return: 2d tensor of cls  output
        """
        self.eval()
        predictions = trainer.predict(model=self, dataloaders=loader, return_predictions=True)
        predictions = cast(List[Dict[int, torch.Tensor]], predictions)
        results = self.get_embeddings(predictions)
        return results


@hydra.main(version_base=HYDRA_VERSION_BASE, config_path="../../../conf", config_name="config")
def start(cfg: DictConfig) -> None:
    trainer: Trainer = instantiate(cfg.SapBertTraining.trainer)
    model: PLSapbertModel = instantiate(cfg.SapBertTraining.model)
    trainer.fit(model)


if __name__ == "__main__":
    start()



import dataclasses
import json
import logging
from abc import ABC, abstractmethod
from collections import defaultdict, Counter
from enum import auto
from typing import (
    cast,
    List,
    Tuple,
    Dict,
    Iterable,
    Set,
    Optional,
    FrozenSet,
    DefaultDict,
    Literal,
    Any,
)

import pandas as pd
from kazu.data.data import (
    EquivalentIdSet,
    EquivalentIdAggregationStrategy,
    SynonymTerm,
    SimpleValue,
    CuratedTerm,
    ParserBehaviour,
    CuratedTermBehaviour,
    AssociatedIdSets,
    GlobalParserActions,
    AutoNameEnum,
    MentionConfidence,
)
from kazu.database.in_memory_db import (
    MetadataDatabase,
    SynonymDatabase,
    NormalisedSynonymStr,
    Idx,
)
from kazu.language.string_similarity_scorers import StringSimilarityScorer
from kazu.ontology_preprocessing.synonym_generation import CombinatorialSynonymGenerator
from kazu.utils.caching import kazu_disk_cache
from kazu.utils.string_normalizer import StringNormalizer
from kazu.utils.utils import PathLike, as_path

# dataframe column keys
DEFAULT_LABEL = "default_label"
IDX = "idx"
SYN = "syn"
MAPPING_TYPE = "mapping_type"
SOURCE = "source"
DATA_ORIGIN = "data_origin"
IdsAndSource = Set[Tuple[str, str]]

logger = logging.getLogger(__name__)


class CurationException(Exception):
    pass


def load_curated_terms(
    path: PathLike,
) -> List[CuratedTerm]:
    """
    Load :class:`kazu.data.data.CuratedTerm`\\ s from a file path.

    :param path: path to json lines file that map to :class:`kazu.data.data.CuratedTerm`
    :return:
    """
    curations_path = as_path(path)
    if curations_path.exists():
        with curations_path.open(mode="r") as jsonlf:
            curations = [CuratedTerm.from_json(line) for line in jsonlf]
    else:
        raise ValueError(f"curations do not exist at {path}")
    return curations


def load_global_actions(
    path: PathLike,
) -> GlobalParserActions:
    """
    Load an instance of :class:`.GlobalParserActions` from a file path.

    :param path: path to a json serialised GlobalParserActions
    :return:
    """
    global_actions_path = as_path(path)
    if global_actions_path.exists():
        with global_actions_path.open(mode="r") as jsonlf:
            global_actions = GlobalParserActions.from_json(json.load(jsonlf))
    else:
        raise ValueError(f"global actions do not exist at {path}")
    return global_actions


class CurationModificationResult(AutoNameEnum):
    ID_SET_MODIFIED = auto()
    SYNONYM_TERM_ADDED = auto()
    SYNONYM_TERM_DROPPED = auto()
    NO_ACTION = auto()


class CurationProcessor:
    """A CurationProcessor is responsible for modifying the set of :class:`.SynonymTerm`\\s produced
    by an :class:`kazu.ontology_preprocessing.base.OntologyParser` with any relevant :class:`.GlobalParserActions` and/or
    :class:`.CuratedTerm` associated with the parser.

    That is to say, this class modifies the raw data produced by a parser with any a posteriori
    observations about the data (such as bad synonyms, mis-mapped terms etc). Is also identifies
    curations that should be used for dictionary based NER.

    This class should be used before instances of :class:`.SynonymTerm`\\s are loaded into the
    internal database representation.
    """

    # curations are applied in the following order
    CURATION_APPLY_ORDER = (
        CuratedTermBehaviour.IGNORE,
        CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING,
        CuratedTermBehaviour.ADD_FOR_LINKING_ONLY,
        CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING,
        CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM,
    )
    _BEHAVIOUR_TO_ORDER_INDEX = {behav: i for i, behav in enumerate(CURATION_APPLY_ORDER)}

    def __init__(
        self,
        parser_name: str,
        entity_class: str,
        global_actions: Optional[GlobalParserActions],
        curations: List[CuratedTerm],
        synonym_terms: Set[SynonymTerm],
    ):
        """

        :param parser_name: name of parser to process
        :param entity_class: name of parser entity_class to process (typically as passed to :class:`kazu.ontology_preprocessing.base.OntologyParser`\\ )
        :param global_actions:
        :param curations:
        :param synonym_terms:
        """
        self.global_actions = global_actions
        self.entity_class = entity_class
        self.parser_name = parser_name
        self._terms_by_term_norm: Dict[NormalisedSynonymStr, SynonymTerm] = {}
        self._terms_by_id: DefaultDict[Idx, Set[SynonymTerm]] = defaultdict(set)
        # used by inherited curations to decide behaviour
        self._curations_by_syn: DefaultDict[str, Set[CuratedTerm]] = defaultdict(set)
        for term in synonym_terms:
            self._update_term_lookups(term, False)
        self.curations = set(curations)

    @classmethod
    def curation_sort_key(cls, curated_term: CuratedTerm):
        """Determines the order curations are processed in.

        We use associated_id_sets as a key, so that any overrides will be
        processed after any original behaviours
        """
        return (
            cls._BEHAVIOUR_TO_ORDER_INDEX[curated_term.behaviour],
            curated_term.associated_id_sets is not None,
            curated_term.curated_synonym,
        )

    def _update_term_lookups(
        self, term: SynonymTerm, override: bool
    ) -> Literal[
        CurationModificationResult.SYNONYM_TERM_ADDED, CurationModificationResult.NO_ACTION
    ]:
        assert term.original_term is None

        safe_to_add = False
        maybe_existing_term = self._terms_by_term_norm.get(term.term_norm)
        if maybe_existing_term is None:
            logger.debug("adding new term %s", term)
            safe_to_add = True
        elif override:
            safe_to_add = True
            logger.debug("overriding existing term %s", maybe_existing_term)
        elif (
            len(
                term.associated_id_sets.symmetric_difference(maybe_existing_term.associated_id_sets)
            )
            > 0
        ):
            logger.warning(
                "conflict on term norms \n%s\n%s\nthe latter will be ignored",
                maybe_existing_term,
                term,
            )
        if safe_to_add:
            self._terms_by_term_norm[term.term_norm] = term
            for equiv_ids in term.associated_id_sets:
                for idx in equiv_ids.ids:
                    self._terms_by_id[idx].add(term)
            return CurationModificationResult.SYNONYM_TERM_ADDED
        else:
            return CurationModificationResult.NO_ACTION

    def _drop_synonym_term(self, synonym: NormalisedSynonymStr) -> None:
        """Remove a synonym term from the database, so that it cannot be used as a linking target.

        :param synonym:
        :return:
        """
        try:
            term_to_remove = self._terms_by_term_norm.pop(synonym)
            for equiv_id_set in term_to_remove.associated_id_sets:
                for idx in equiv_id_set.ids:
                    terms_by_id = self._terms_by_id.get(idx)
                    if terms_by_id is not None:
                        terms_by_id.remove(term_to_remove)
            logger.debug(
                "successfully dropped %s from database for %s",
                synonym,
                self.entity_class,
            )
        except KeyError:
            logger.warning(
                "tried to drop %s from database, but key doesn't exist for %s",
                synonym,
                self.parser_name,
            )

    def _drop_id_from_all_synonym_terms(self, id_to_drop: Idx) -> Counter:
        """Remove a given id from all :class:`.SynonymTerm`\\ s.

        Drop any :class:`.SynonymTerm`\\ s with no remaining ID after removal.

        :param id_to_drop:
        :return: counter of :class:`.CurationModificationResult`
        """

        terms_to_modify = self._terms_by_id.get(id_to_drop, set())
        counter = Counter(
            self._drop_id_from_synonym_term(id_to_drop=id_to_drop, term_to_modify=term_to_modify)
            for term_to_modify in set(terms_to_modify)
        )

        return counter

    def _drop_id_from_synonym_term(
        self, id_to_drop: Idx, term_to_modify: SynonymTerm
    ) -> Literal[
        CurationModificationResult.ID_SET_MODIFIED,
        CurationModificationResult.SYNONYM_TERM_DROPPED,
        CurationModificationResult.NO_ACTION,
    ]:
        """Remove an id from a given :class:`.SynonymTerm`\\ .

        :param id_to_drop:
        :param term_to_modify:
        :return:
        """
        new_assoc_id_frozenset = self._drop_id_from_associated_id_sets(
            id_to_drop, term_to_modify.associated_id_sets
        )
        if len(new_assoc_id_frozenset.symmetric_difference(term_to_modify.associated_id_sets)) == 0:
            return CurationModificationResult.NO_ACTION
        else:
            return self._modify_or_drop_synonym_term_after_id_set_change(
                new_associated_id_sets=new_assoc_id_frozenset, synonym_term=term_to_modify
            )

    def _drop_id_from_associated_id_sets(
        self, id_to_drop: Idx, associated_id_sets: AssociatedIdSets
    ) -> AssociatedIdSets:
        """Remove an id from a :class:`.AssociatedIdSets`\\ .

        :param id_to_drop:
        :param associated_id_sets:
        :return:
        """
        new_assoc_id_set = set()
        for equiv_id_set in associated_id_sets:
            if id_to_drop in equiv_id_set.ids:
                updated_ids_and_source = frozenset(
                    id_tup for id_tup in equiv_id_set.ids_and_source if id_tup[0] != id_to_drop
                )
                if len(updated_ids_and_source) > 0:
                    updated_equiv_id_set = EquivalentIdSet(updated_ids_and_source)
                    new_assoc_id_set.add(updated_equiv_id_set)
            else:
                new_assoc_id_set.add(equiv_id_set)
        new_assoc_id_frozenset = frozenset(new_assoc_id_set)
        return new_assoc_id_frozenset

    def _modify_or_drop_synonym_term_after_id_set_change(
        self, new_associated_id_sets: AssociatedIdSets, synonym_term: SynonymTerm
    ) -> Literal[
        CurationModificationResult.ID_SET_MODIFIED, CurationModificationResult.SYNONYM_TERM_DROPPED
    ]:
        """Modifies or drops a :class:`.SynonymTerm` after a :class:`.AssociatedIdSets` has
        changed.

        :param new_associated_id_sets:
        :param synonym_term:
        :return:
        """
        result: Literal[
            CurationModificationResult.ID_SET_MODIFIED,
            CurationModificationResult.SYNONYM_TERM_DROPPED,
        ]
        if len(new_associated_id_sets) > 0:
            if new_associated_id_sets == synonym_term.associated_id_sets:
                raise ValueError(
                    "function called inappropriately where the id sets haven't changed. This"
                    " has failed as it will otherwise modify the value of aggregated_by, when"
                    " nothing has changed"
                )
            new_term = dataclasses.replace(
                synonym_term,
                associated_id_sets=new_associated_id_sets,
                aggregated_by=EquivalentIdAggregationStrategy.MODIFIED_BY_CURATION,
            )
            add_result = self._update_term_lookups(new_term, True)
            assert add_result is CurationModificationResult.SYNONYM_TERM_ADDED
            result = CurationModificationResult.ID_SET_MODIFIED
        else:
            # if there are no longer any id sets associated with the record, remove it completely
            self._drop_synonym_term(synonym_term.term_norm)
            result = CurationModificationResult.SYNONYM_TERM_DROPPED
        return result

    def export_curations_and_final_terms(
        self,
    ) -> Tuple[List[CuratedTerm], Set[SynonymTerm]]:
        """Perform any updates required to the synonym terms as specified in the curations/global
        actions.

        The returned :class:`.CuratedTerm`\\s can be used for Dictionary based NER, whereas the
        returned :class:`.SynonymTerm`\\s can be loaded into the internal database for linking.

        :return:
        """
        self._process_global_actions()
        return list(self._process_curations()), set(self._terms_by_term_norm.values())

    def _process_curations(self) -> Iterable[CuratedTerm]:
        safe_curations = self.fix_conflicts_in_curations(self.curations)
        for curation in sorted(safe_curations, key=self.curation_sort_key):
            curation = self._process_curation_action(curation)
            yield curation

    def fix_conflicts_in_curations(self, curations: Set[CuratedTerm]) -> Set[CuratedTerm]:
        """Check to see if a list of curations contain conflicts.

        Conflicts can occur for the following reasons:

        1) If two or more curations normalise to the same NormalisedSynonymStr,
           but have different behaviours.

        2) If two or more curations normalise to the same NormalisedSynonymStr,
           but have different associated ID sets specified, such that one would
           override the other.

        3) If two or more curations have conflicting values for case sensitivity and
           mention_confidence. E.g. A case-insensitive curation cannot have a higher
           mention confidence value than a case-sensitive one for the same synonym.


        :param curations:
        :return: safe curations set
        """

        curations_by_term_norm = defaultdict(set)
        curations_by_syn_lower = defaultdict(set)
        for curation in curations:
            curations_by_term_norm[curation.term_norm_for_linking(self.entity_class)].add(curation)
            curations_by_syn_lower[curation.curated_synonym.lower()].add(curation)

        all_remove = set()
        for potential_conflict_set in curations_by_term_norm.values():
            if len(potential_conflict_set) > 1:
                to_add, to_remove = self.resolve_behaviour_conflicts(potential_conflict_set)
                all_remove.update(to_remove)
                curations.update(to_add)

        for potential_conflict_set in curations_by_syn_lower.values():
            potential_conflict_set.difference_update(all_remove)
            if len(potential_conflict_set) > 1:
                to_add, to_remove = self.resolve_case_conflicts(potential_conflict_set)
                all_remove.update(to_remove)
                curations.update(to_add)

        curations.difference_update(all_remove)
        return curations

    def resolve_behaviour_conflicts(
        self, curations: Set[CuratedTerm]
    ) -> Tuple[Set[CuratedTerm], Set[CuratedTerm]]:

        curations_by_syn_lower = defaultdict(set)
        potentially_conflicting_behaviours = set()
        potentially_conflicting_id_sets = set()
        source_curations = set()
        inherited_curations = set()
        for curation in curations:
            if curation.source_term is None:
                source_curations.add(curation)
                if curation.behaviour in {
                    CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING,
                    CuratedTermBehaviour.ADD_FOR_LINKING_ONLY,
                    CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING,
                }:
                    potentially_conflicting_behaviours.add(curation.behaviour)
                if curation.associated_id_sets is not None:
                    potentially_conflicting_id_sets.add(curation.associated_id_sets)
            else:
                inherited_curations.add(curation)
            curations_by_syn_lower[curation.curated_synonym.lower()].add(curation)

        if len(potentially_conflicting_behaviours) > 1:
            resolved_behaviour = (
                CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING
                if CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING
                in potentially_conflicting_behaviours
                else CuratedTermBehaviour.IGNORE
            )
            logger.warning(
                "conflicting behaviours detected. The following source curations will be %s\n%s",
                resolved_behaviour.name,
                "\n\n".join(curation.to_json() for curation in source_curations) + "\n\n",
            )
            if len(inherited_curations) > 0:
                logger.warning(
                    "conflicting behaviours detected. The following inherited curations will be %s\n%s",
                    CuratedTermBehaviour.IGNORE.name,
                    "\n\n".join(curation.to_json() for curation in inherited_curations) + "\n\n",
                )
            return (
                set(
                    dataclasses.replace(curation, behaviour=resolved_behaviour)
                    for curation in source_curations
                ).union(
                    dataclasses.replace(curation, behaviour=CuratedTermBehaviour.IGNORE)
                    for curation in inherited_curations
                ),
                curations,
            )
        if len(potentially_conflicting_id_sets) > 1:
            raise CurationException(
                "conflicting id sets detected in curations. Please fix the below curations\n%s",
                "\n\n".join(curation.to_json() for curation in source_curations)
                + "\n\n"
                + "\n\n".join(curation.to_json() for curation in inherited_curations)
                + "\n\n",
            )
        return set(), set()

    def resolve_case_conflicts(
        self, curations: Set[CuratedTerm]
    ) -> Tuple[Set[CuratedTerm], Set[CuratedTerm]]:

        to_add: Set[CuratedTerm] = set()
        to_remove: Set[CuratedTerm] = set()
        cs_conf = set()
        ci_conf = set()
        cs_lookup: DefaultDict[str, Tuple[Set[CuratedTerm], Set[MentionConfidence]]] = defaultdict(
            lambda: (set(), set())
        )
        for curation in curations:
            if curation.behaviour in {
                CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM,
                CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING,
            }:
                if curation.case_sensitive:
                    cs_conf.add(curation.mention_confidence)
                    cs_lookup[curation.curated_synonym][0].add(curation)
                    cs_lookup[curation.curated_synonym][1].add(curation.mention_confidence)
                else:
                    ci_conf.add(curation.mention_confidence)

        if len(ci_conf.union(cs_conf)) == 1:
            logger.debug("curations OK \n %s", curations)
        elif len(ci_conf) > 1 or (
            len(ci_conf) > 0 and len(cs_conf) > 0 and min(ci_conf) < min(cs_conf)
        ):
            # set all to CI min
            target_conf = min(ci_conf)
            logger.warning(
                "conflict detected in case sensitivity/confidence combination for the following curations. Affected curations will adopt the conservative confidence value %s,%s",
                target_conf.name,
                "\n".join(curation.to_json() for curation in curations) + "\n\n",
            )
            to_add.update(
                dataclasses.replace(curation, mention_confidence=target_conf)
                for curation in curations
            )
            to_remove.update(curations)
        elif len(cs_conf) > 1:
            for curations_cs, cs_confs_by_exact_match in cs_lookup.values():
                if len(cs_confs_by_exact_match) > 1:
                    target_conf = min(cs_confs_by_exact_match)
                    logger.warning(
                        "conflict detected in case sensitivity/confidence combination for the following curations. Affected curations will adopt the conservative confidence value %s,%s",
                        target_conf.name,
                        "\n".join(curation.to_json() for curation in curations_cs) + "\n\n",
                    )
                    to_add.update(
                        dataclasses.replace(curation, mention_confidence=target_conf)
                        for curation in curations_cs
                    )
                    to_remove.update(curations_cs)

        return to_add, to_remove

    def _process_curation_action(self, curation: CuratedTerm) -> CuratedTerm:

        if curation.source_term is None:
            self._curations_by_syn[curation.curated_synonym].add(curation)
        term_norm = curation.term_norm_for_linking(self.entity_class)
        if curation.behaviour is CuratedTermBehaviour.IGNORE:
            logger.debug("curation ignored: %s for %s", curation, self.parser_name)
        elif curation.behaviour is CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM:
            assert curation.source_term is not None, curation
            logger.debug(
                "curation inherits linking behaviour from %s for %s",
                curation.source_term,
                self.parser_name,
            )
            source_curations = self._curations_by_syn.get(curation.source_term)
            if source_curations is None:
                logger.warning(
                    "no source curation found for %s for %s. The curation will be ignored",
                    curation.source_term,
                    self.parser_name,
                )
                return dataclasses.replace(curation, behaviour=CuratedTermBehaviour.IGNORE)
            elif len(source_curations) > 1:
                logger.warning(
                    "multiple source curations found for %s for %s. The curation will be ignored",
                    curation.source_term,
                    self.parser_name,
                )
                return dataclasses.replace(curation, behaviour=CuratedTermBehaviour.IGNORE)
        elif curation.behaviour is CuratedTermBehaviour.DROP_SYNONYM_TERM_FOR_LINKING:
            self._drop_synonym_term(term_norm)
        elif curation.behaviour is CuratedTermBehaviour.ADD_FOR_LINKING_ONLY:
            self._attempt_to_add_database_entry_for_curation(
                curation_associated_id_set=curation.associated_id_sets,
                curated_synonym=curation.curated_synonym,
                curation_term_norm=term_norm,
            )

        elif curation.behaviour is CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING:
            self._attempt_to_add_database_entry_for_curation(
                curation_associated_id_set=curation.associated_id_sets,
                curated_synonym=curation.curated_synonym,
                curation_term_norm=term_norm,
            )
            term_for_this_curation = self._terms_by_term_norm.get(term_norm)
            if term_for_this_curation is None:
                logger.warning(
                    "CuratedTerm %s is invalid: "
                    "requires an identifier but none was found. It may have been removed by another curation, or not exist in the underlying data sourcee.",
                    curation,
                )
                return dataclasses.replace(curation, behaviour=CuratedTermBehaviour.IGNORE)
        else:
            raise ValueError(f"unknown behaviour for parser {self.parser_name}, {curation}")
        return curation

    def _process_global_actions(self) -> None:
        if self.global_actions is None:
            return None

        override_curations_by_id = defaultdict(set)
        for curation in self.curations:
            if curation.associated_id_sets is not None:
                for equiv_id_set in curation.associated_id_sets:
                    for idx in equiv_id_set.ids:
                        override_curations_by_id[idx].add(curation)

        for action in self.global_actions.parser_behaviour(self.parser_name):
            if action.behaviour is ParserBehaviour.DROP_IDS_FROM_PARSER:
                ids = action.parser_to_target_id_mappings[self.parser_name]
                for idx in ids:
                    counter_this_idx = self._drop_id_from_all_synonym_terms(idx)
                    if (
                        counter_this_idx[CurationModificationResult.ID_SET_MODIFIED]
                        + counter_this_idx[CurationModificationResult.SYNONYM_TERM_DROPPED]
                        == 0
                    ):
                        logger.warning("failed to drop %s from %s", idx, self.parser_name)
                    else:
                        logger.debug(
                            "dropped ID %s from %s. SynonymTerm modified count: %s, SynonymTerm dropped count: %s",
                            idx,
                            self.parser_name,
                            counter_this_idx[CurationModificationResult.ID_SET_MODIFIED],
                            counter_this_idx[CurationModificationResult.SYNONYM_TERM_DROPPED],
                        )

                        for override_curation_to_modify in set(
                            override_curations_by_id.get(idx, set())
                        ):
                            assert override_curation_to_modify.associated_id_sets is not None
                            new_associated_id_sets = self._drop_id_from_associated_id_sets(
                                idx, override_curation_to_modify.associated_id_sets
                            )
                            if len(new_associated_id_sets) == 0:

                                self.curations.remove(override_curation_to_modify)
                                override_curations_by_id[idx].remove(override_curation_to_modify)
                                logger.debug(
                                    "removed curation %s because of global action",
                                    override_curation_to_modify,
                                )
                            elif (
                                new_associated_id_sets
                                != override_curation_to_modify.associated_id_sets
                            ):
                                self.curations.remove(override_curation_to_modify)
                                override_curations_by_id[idx].remove(override_curation_to_modify)
                                mod_curation = dataclasses.replace(
                                    override_curation_to_modify,
                                    associated_id_sets=new_associated_id_sets,
                                )
                                self.curations.add(mod_curation)
                                override_curations_by_id[idx].add(mod_curation)
                                logger.debug(
                                    "modified curation %s to %s because of global action",
                                    override_curation_to_modify,
                                    mod_curation,
                                )

            else:
                raise ValueError(f"unknown behaviour for parser {self.parser_name}, {action}")
        return None

    def _attempt_to_add_database_entry_for_curation(
        self,
        curation_term_norm: NormalisedSynonymStr,
        curation_associated_id_set: Optional[AssociatedIdSets],
        curated_synonym: str,
    ) -> Literal[
        CurationModificationResult.SYNONYM_TERM_ADDED, CurationModificationResult.NO_ACTION
    ]:
        """
        Create a new :class:`~kazu.data.data.SynonymTerm` for the database, or return an existing
        matching one if already present.

        Notes:

        If a term_norm already exists in self._terms_by_term_norm that matches 'curation_term_norm',
        this method will check to see if the 'curation_associated_id_set' matches the existing terms
        :class:`.AssociatedIdSets`\\.

        If so, no action will be taken. If not, a warning will be logged as adding it
        will cause irregularities in the database.

        If the term_norm does not exist, this method will create a new :class:`~kazu.data.data.SynonymTerm`
        with the provided :class:`.AssociatedIdSets`\\.

        :param curation_term_norm:
        :param curation_associated_id_set:
        :param curated_synonym:
        :return:
        """
        log_prefix = "%(parser_name)s attempting to create synonym term for <%(synonym)s> term_norm: <%(term_norm)s> IDs: %(ids)s}"
        log_formatting_dict: Dict[str, Any] = {
            "parser_name": self.parser_name,
            "synonym": curated_synonym,
            "term_norm": curation_term_norm,
            "ids": curation_associated_id_set,
        }

        # look up the term norm in the db
        maybe_existing_synonym_term = self._terms_by_term_norm.get(curation_term_norm)
        if curation_associated_id_set is None and maybe_existing_synonym_term is not None:
            logger.debug(
                log_prefix
                + " but no associated id set provided, so term will inherit the parser defaults",
                log_formatting_dict,
            )
            return CurationModificationResult.NO_ACTION
        elif curation_associated_id_set is None and maybe_existing_synonym_term is None:
            logger.error(
                log_prefix
                + " but term_norm <%(term_norm)s> does not exist in synonym database."
                + " Since no id set was provided, no entry can be created",
                log_formatting_dict,
            )
            return CurationModificationResult.NO_ACTION

        # curation_associated_id_set is implicitly not None
        assert curation_associated_id_set is not None
        if len(curation_associated_id_set) == 0:
            logger.debug(
                "all ids removed by global action for %s,  Parser name: %s",
                curated_synonym,
                self.parser_name,
            )
            return CurationModificationResult.NO_ACTION

        if maybe_existing_synonym_term is not None:
            log_formatting_dict["existing_id_set"] = maybe_existing_synonym_term.associated_id_sets

            if (
                len(
                    curation_associated_id_set.symmetric_difference(
                        maybe_existing_synonym_term.associated_id_sets
                    )
                )
                == 0
            ):
                logger.debug(
                    log_prefix
                    + " but term_norm <%(term_norm)s> already exists in synonym database."
                    + "since this SynonymTerm matches the id_set, no action is required. %(existing_id_set)s",
                    log_formatting_dict,
                )
                return CurationModificationResult.NO_ACTION
            else:
                logger.debug(
                    log_prefix
                    + " . Will remove existing term_norm <%(term_norm)s> as an ID set override has been specified",
                    log_formatting_dict,
                )

        # no term exists, or we want to override so one will be made
        assert curation_associated_id_set is not None
        for equiv_id_set in set(curation_associated_id_set):
            for idx in equiv_id_set.ids:
                if idx not in self._terms_by_id:
                    curation_associated_id_set = self._drop_id_from_associated_id_sets(
                        id_to_drop=idx, associated_id_sets=curation_associated_id_set
                    )
                    logger.warning(
                        "Attempted to add term containing %s but this id does not exist in the parser and will be ignored",
                        idx,
                    )
        if len(curation_associated_id_set) > 0:
            is_symbolic = StringNormalizer.classify_symbolic(curated_synonym, self.entity_class)
            new_term = SynonymTerm(
                term_norm=curation_term_norm,
                terms=frozenset((curated_synonym,)),
                is_symbolic=is_symbolic,
                mapping_types=frozenset(("kazu_curated",)),
                associated_id_sets=curation_associated_id_set,
                parser_name=self.parser_name,
                aggregated_by=EquivalentIdAggregationStrategy.MODIFIED_BY_CURATION,
            )
            return self._update_term_lookups(new_term, True)
        else:
            return CurationModificationResult.NO_ACTION


class OntologyParser(ABC):
    """
    Parse an ontology (or similar) into a set of outputs suitable for NLP entity linking.
    Implementations should have a class attribute 'name' to something suitably representative.
    The key method is parse_to_dataframe, which should convert an input source to a dataframe suitable
    for further processing.

    The other important method is find_kb. This should parse an ID string (if required) and return the underlying
    source. This is important for composite resources that contain identifiers from different seed sources

    Generally speaking, when parsing a data source, synonyms that are symbolic (as determined by
    the StringNormalizer) that refer to more than one id are more likely to be ambiguous. Therefore,
    we assume they refer to unique concepts (e.g. COX 1 could be 'ENSG00000095303' OR
    'ENSG00000198804', and thus they will yield multiple instances of EquivalentIdSet. Non symbolic
    synonyms (i.e. noun phrases) are far less likely to refer to distinct entities, so we might
    want to merge the associated ID's non-symbolic ambiguous synonyms into a single EquivalentIdSet.
    The result of StringNormalizer.is_symbolic forms the is_symbolic parameter to .score_and_group_ids.

    If the underlying knowledgebase contains more than one entity type, muliple parsers should be
    implemented, subsetting accordingly (e.g. MEDDRA_DISEASE, MEDDRA_DIAGNOSTIC).
    """

    # the synonym table should have these (and only these columns)
    all_synonym_column_names = [IDX, SYN, MAPPING_TYPE]
    # the metadata table should have at least these columns (note, IDX will become the index)
    minimum_metadata_column_names = [DEFAULT_LABEL, DATA_ORIGIN]

    def __init__(
        self,
        in_path: str,
        entity_class: str,
        name: str,
        string_scorer: Optional[StringSimilarityScorer] = None,
        synonym_merge_threshold: float = 0.70,
        data_origin: str = "unknown",
        synonym_generator: Optional[CombinatorialSynonymGenerator] = None,
        curations_path: Optional[str] = None,
        global_actions: Optional[GlobalParserActions] = None,
    ):
        """
        :param in_path: Path to some resource that should be processed (e.g. owl file, db config, tsv etc)
        :param entity_class: The entity class to associate with this parser throughout the pipeline.
            Also used in the parser when calling StringNormalizer to determine the class-appropriate behaviour.
        :param name: A string to represent a parser in the overall pipeline. Should be globally unique
        :param string_scorer: Optional protocol of StringSimilarityScorer.  Used for resolving ambiguous symbolic
            synonyms via similarity calculation of the default label associated with the conflicted labels. If no
            instance is provided, all synonym conflicts will be assumed to refer to different concepts. This is not
            recommended!
        :param synonym_merge_threshold: similarity threshold to trigger a merge of conflicted synonyms into a single
            EquivalentIdSet. See docs for score_and_group_ids for further details
        :param data_origin: The origin of this dataset - e.g. HGNC release 2.1, MEDDRA 24.1 etc. Note, this is different
            from the parser.name, as is used to identify the origin of a mapping back to a data source
        :param synonym_generator: optional CombinatorialSynonymGenerator. Used to generate synonyms for dictionary
            based NER matching
        :param curations_path: path to jsonl file of curations to apply to the parser
        """

        self.in_path = in_path
        self.entity_class = entity_class
        self.name = name
        if string_scorer is None:
            logger.warning(
                "no string scorer configured for %s. Synonym resolution disabled.", self.name
            )
        self.string_scorer = string_scorer
        self.synonym_merge_threshold = synonym_merge_threshold
        self.data_origin = data_origin
        self.synonym_generator = synonym_generator
        self.curations_path = curations_path
        self.global_actions = global_actions
        self.parsed_dataframe: Optional[pd.DataFrame] = None
        self.metadata_db = MetadataDatabase()
        self.synonym_db = SynonymDatabase()

    @abstractmethod
    def find_kb(self, string: str) -> str:
        """
        split an IDX somehow to find the ontology SOURCE reference

        :param string: the IDX string to process
        :return:
        """
        pass

    def resolve_synonyms(self, synonym_df: pd.DataFrame) -> Set[SynonymTerm]:

        result = set()
        synonym_df["syn_norm"] = synonym_df[SYN].apply(
            StringNormalizer.normalize, entity_class=self.entity_class
        )

        for i, row in (
            synonym_df[["syn_norm", SYN, IDX, MAPPING_TYPE]]
            .groupby(["syn_norm"])
            .agg(set)
            .reset_index()
            .iterrows()
        ):

            syn_set = row[SYN]
            mapping_type_set: FrozenSet[str] = frozenset(row[MAPPING_TYPE])
            syn_norm = row["syn_norm"]
            if len(syn_set) > 1:
                logger.debug("normaliser has merged %s into a single term: %s", syn_set, syn_norm)

            is_symbolic = all(
                StringNormalizer.classify_symbolic(x, self.entity_class) for x in syn_set
            )

            ids: Set[str] = row[IDX]
            ids_and_source = set(
                (
                    idx,
                    self.find_kb(idx),
                )
                for idx in ids
            )
            associated_id_sets, agg_strategy = self.score_and_group_ids(
                ids_and_source, is_symbolic, syn_set
            )

            synonym_term = SynonymTerm(
                term_norm=syn_norm,
                terms=frozenset(syn_set),
                is_symbolic=is_symbolic,
                mapping_types=mapping_type_set,
                associated_id_sets=associated_id_sets,
                parser_name=self.name,
                aggregated_by=agg_strategy,
            )

            result.add(synonym_term)

        return result

    def score_and_group_ids(
        self,
        ids_and_source: IdsAndSource,
        is_symbolic: bool,
        original_syn_set: Set[str],
    ) -> Tuple[AssociatedIdSets, EquivalentIdAggregationStrategy]:
        """
        for a given data source, one normalised synonym may map to one or more id. In some cases, the ID may be
        duplicate/redundant (e.g. there are many chembl ids for paracetamol). In other cases, the ID may refer to
        distinct concepts (e.g. COX 1 could be 'ENSG00000095303' OR 'ENSG00000198804').


        Since synonyms from data sources are confused in such a manner, we need to decide some way to cluster them into
        a single SynonymTerm concept, which in turn is a container for one or more EquivalentIdSet (depending on
        whether the concept is ambiguous or not)

        The job of score_and_group_ids is to determine how many EquivalentIdSet's for a given set of ids should be
        produced.

        The default algorithm (which can be overridden by concrete parser implementations) works as follows:

        1. If no StringScorer is configured, create an EquivalentIdSet for each id (strategy NO_STRATEGY -
           not recommended)
        2. If only one ID is referenced, or the associated normalised synonym string is not symbolic, group the
           ids into a single EquivalentIdSet (strategy UNAMBIGUOUS)
        3. otherwise, compare the default label associated with each ID to every other default label. If it's above
           self.synonym_merge_threshold, merge into one EquivalentIdSet, if not, create a new one

        recommendation: Use the SapbertStringSimilarityScorer for comparison

        IMPORTANT NOTE: any calls to this method requires the metadata DB to be populated, as this is the store of
        DEFAULT_LABEL

        :param ids_and_source: ids to determine appropriate groupings of, and their associated sources
        :param is_symbolic: is the underlying synonym symbolic?
        :param original_syn_set: original synonyms associated with ids
        :return:
        """
        if self.string_scorer is None:
            # the NO_STRATEGY aggregation strategy assumes all synonyms are ambiguous
            return (
                frozenset(
                    EquivalentIdSet(ids_and_source=frozenset((single_id_and_source,)))
                    for single_id_and_source in ids_and_source
                ),
                EquivalentIdAggregationStrategy.NO_STRATEGY,
            )
        else:

            if len(ids_and_source) == 1:
                return (
                    frozenset((EquivalentIdSet(ids_and_source=frozenset(ids_and_source)),)),
                    EquivalentIdAggregationStrategy.UNAMBIGUOUS,
                )

            if not is_symbolic:
                return (
                    frozenset((EquivalentIdSet(ids_and_source=frozenset(ids_and_source)),)),
                    EquivalentIdAggregationStrategy.MERGED_AS_NON_SYMBOLIC,
                )
            else:
                # use similarity to group ids into EquivalentIdSets

                DefaultLabels = Set[str]
                id_list: List[Tuple[IdsAndSource, DefaultLabels]] = []
                for id_and_source_tuple in ids_and_source:
                    default_label = cast(
                        str,
                        self.metadata_db.get_by_idx(self.name, id_and_source_tuple[0])[
                            DEFAULT_LABEL
                        ],
                    )
                    most_similar_id_set = None
                    best_score = 0.0
                    for id_and_default_label_set in id_list:
                        sim = max(
                            self.string_scorer(default_label, other_label)
                            for other_label in id_and_default_label_set[1]
                        )
                        if sim > self.synonym_merge_threshold and sim > best_score:
                            most_similar_id_set = id_and_default_label_set
                            best_score = sim

                    # for the first label, the above for loop is a no-op as id_sets is empty
                    # and the below if statement will be true.
                    # After that, it will be True if the id under consideration should not
                    # merge with any existing group and should get its own EquivalentIdSet
                    if not most_similar_id_set:
                        id_list.append(
                            (
                                {id_and_source_tuple},
                                {default_label},
                            )
                        )
                    else:
                        most_similar_id_set[0].add(id_and_source_tuple)
                        most_similar_id_set[1].add(default_label)

                return (
                    frozenset(
                        EquivalentIdSet(ids_and_source=frozenset(ids_and_source))
                        for ids_and_source, _ in id_list
                    ),
                    EquivalentIdAggregationStrategy.RESOLVED_BY_SIMILARITY,
                )

    def _parse_df_if_not_already_parsed(self):
        if self.parsed_dataframe is None:
            self.parsed_dataframe = self.parse_to_dataframe()
            self.parsed_dataframe[DATA_ORIGIN] = self.data_origin
            self.parsed_dataframe[IDX] = self.parsed_dataframe[IDX].astype(str)
            self.parsed_dataframe.loc[
                pd.isnull(self.parsed_dataframe[DEFAULT_LABEL]), DEFAULT_LABEL
            ] = self.parsed_dataframe[IDX]

    @kazu_disk_cache.memoize(ignore={0})
    def export_metadata(self, parser_name: str) -> Dict[str, Dict[str, SimpleValue]]:
        """Export the metadata from the ontology.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too much
            memory)
        :return: {idx:{metadata_key:metadata_value}}
        """
        self._parse_df_if_not_already_parsed()
        assert isinstance(self.parsed_dataframe, pd.DataFrame)
        metadata_columns = self.parsed_dataframe.columns
        metadata_columns.drop([MAPPING_TYPE, SYN])
        metadata_df = self.parsed_dataframe[metadata_columns]
        metadata_df = metadata_df.drop_duplicates(subset=[IDX]).dropna(axis=0)
        metadata_df.set_index(inplace=True, drop=True, keys=IDX)
        assert set(OntologyParser.minimum_metadata_column_names).issubset(metadata_df.columns)
        metadata = metadata_df.to_dict(orient="index")
        return cast(Dict[str, Dict[str, SimpleValue]], metadata)

    def process_curations(
        self, terms: Set[SynonymTerm]
    ) -> Tuple[Optional[List[CuratedTerm]], Set[SynonymTerm]]:
        if self.curations_path is None and self.synonym_generator is not None:
            logger.warning(
                "%s is configured to use synonym generators. This may result in noisy NER performance.",
                self.name,
            )
            (
                original_curations,
                generated_curations,
            ) = self.generate_curations_from_synonym_generators(terms)
            curations = original_curations + generated_curations
        elif (
            self.curations_path is None
        ):  # implies self.synonym_generator is None as failed the if above
            logger.warning(
                "%s is configured to use raw ontology synonyms. This may result in noisy NER performance.",
                self.name,
            )
            curations = [
                curation
                for term in terms
                for curation in self.synonym_term_to_putative_curation(term)
            ]
        else:
            assert self.curations_path is not None
            logger.info(
                "%s is configured to use curations. Synonym generation will be ignored",
                self.name,
            )
            curations = load_curated_terms(self.curations_path)

        curation_processor = CurationProcessor(
            global_actions=self.global_actions,
            curations=curations,
            parser_name=self.name,
            entity_class=self.entity_class,
            synonym_terms=terms,
        )
        return curation_processor.export_curations_and_final_terms()

    @kazu_disk_cache.memoize(ignore={0})
    def export_synonym_terms(self, parser_name: str) -> Set[SynonymTerm]:
        """Export :class:`.SynonymTerm` from the parser.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too much
            memory)
        :return:
        """
        self._parse_df_if_not_already_parsed()
        assert isinstance(self.parsed_dataframe, pd.DataFrame)
        # ensure correct order
        syn_df = self.parsed_dataframe[self.all_synonym_column_names].copy()
        syn_df = syn_df.dropna(subset=[SYN])
        syn_df[SYN] = syn_df[SYN].apply(str.strip)
        syn_df.drop_duplicates(subset=self.all_synonym_column_names)
        assert set(OntologyParser.all_synonym_column_names).issubset(syn_df.columns)
        synonym_terms = self.resolve_synonyms(synonym_df=syn_df)
        return synonym_terms

    def generate_synonyms(
        self, original_synonym_data: Set[SynonymTerm]
    ) -> Tuple[Set[SynonymTerm], Set[SynonymTerm]]:
        """Generate synonyms based on configured synonym generator.

        :param original_synonym_data:
        :return: first set is original terms, second are generated terms
        """
        generated_synonym_data = set()
        if self.synonym_generator:
            generated_synonym_data = self.synonym_generator(original_synonym_data)
        logger.info(
            f"{len(original_synonym_data)} original synonyms and {len(generated_synonym_data)} generated synonyms produced"
        )
        return original_synonym_data, generated_synonym_data

    def generate_curations_from_synonym_generators(
        self, synonym_terms: Set[SynonymTerm]
    ) -> Tuple[List[CuratedTerm], List[CuratedTerm]]:
        original_terms, generated_terms = self.generate_synonyms(synonym_terms)
        original_curations = [
            curation
            for term in original_terms
            for curation in self.synonym_term_to_putative_curation(term)
        ]
        generated_curations = [
            curation
            for term in generated_terms
            for curation in self.synonym_term_to_putative_curation(term)
        ]

        return original_curations, generated_curations

    def synonym_term_to_putative_curation(self, term: SynonymTerm) -> Iterable[CuratedTerm]:
        """Convert a :class:`.SynonymTerm`\\ s to curations to use for dictionary based NER.

        This is used when curations are not provided to the parser.

        Can handle either 'original' or 'generated' :class:`.SynonymTerm`\\ s.

        :param term:
        :return:
        """
        for term_str in term.terms:
            if term.original_term is None:
                behaviour = CuratedTermBehaviour.ADD_FOR_NER_AND_LINKING
            else:
                behaviour = CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM
            is_symbolic = StringNormalizer.classify_symbolic(term_str, self.entity_class)
            conf = MentionConfidence.POSSIBLE if is_symbolic else MentionConfidence.PROBABLE
            yield CuratedTerm(
                curated_synonym=term_str,
                mention_confidence=conf,
                case_sensitive=is_symbolic,
                behaviour=behaviour,
                source_term=term.original_term,
            )

    @kazu_disk_cache.memoize(ignore={0})
    def _populate_databases(
        self, parser_name: str
    ) -> Tuple[Optional[List[CuratedTerm]], Dict[str, Dict[str, SimpleValue]], Set[SynonymTerm]]:
        """Disk cacheable method that populates all databases.

        :param parser_name: name of this parser. Required for correct operation of cache
            (Note, we cannot pass self to the disk cache as the constructor consumes too much
            memory)
        :return:
        """
        logger.info("populating database for %s from source", self.name)
        metadata = self.export_metadata(self.name)
        # metadata db needs to be populated before call to export_synonym_terms
        self.metadata_db.add_parser(self.name, metadata)
        intermediate_synonym_terms = self.export_synonym_terms(self.name)
        maybe_ner_curations, final_syn_terms = self.process_curations(intermediate_synonym_terms)
        self.parsed_dataframe = None  # clear the reference to save memory

        self.synonym_db.add(self.name, final_syn_terms)
        return maybe_ner_curations, metadata, final_syn_terms

    def populate_databases(
        self, force: bool = False, return_curations: bool = False
    ) -> Optional[List[CuratedTerm]]:
        """Populate the databases with the results of the parser.

        Also calculates the term norms associated with
        any curations (if provided) which can then be used for Dictionary based NER

        :param force: do not use the cache for the ontology parser
        :param return_curations: should processed curations be returned?
        :return: curations if required
        """
        if self.name in self.synonym_db.loaded_parsers and not force and not return_curations:
            logger.debug("parser %s already loaded.", self.name)
            return None

        cache_key = self._populate_databases.__cache_key__(self, self.name)

        if force:
            kazu_disk_cache.delete(self.export_metadata.__cache_key__(self, self.name))
            kazu_disk_cache.delete(self.export_synonym_terms.__cache_key__(self, self.name))
            kazu_disk_cache.delete(cache_key)

        maybe_curations, metadata, final_syn_terms = self._populate_databases(self.name)

        if self.name not in self.synonym_db.loaded_parsers:
            logger.info("populating database for %s from cache", self.name)
            self.metadata_db.add_parser(self.name, metadata)
            self.synonym_db.add(self.name, final_syn_terms)

        return maybe_curations if return_curations else None

    @abstractmethod
    def parse_to_dataframe(self) -> pd.DataFrame:
        """
        implementations should override this method, returning a 'long, thin' pd.DataFrame of at least the following
        columns:


        [IDX, DEFAULT_LABEL, SYN, MAPPING_TYPE]

        IDX: the ontology id
        DEFAULT_LABEL: the preferred label
        SYN: a synonym of the concept
        MAPPING_TYPE: the type of mapping from default label to synonym - e.g. xref, exactSyn etc. Usually defined by the ontology

        Note: It is the responsibility of the implementation of parse_to_dataframe to add default labels as synonyms.

        Any 'extra' columns will be added to the :class:`~kazu.database.in_memory_db.MetadataDatabase` as metadata fields for the
        given id in the relevant ontology.
        """
        pass



import json
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Iterable, Set, Tuple, DefaultDict, Dict, List

import requests
from requests.adapters import HTTPAdapter, Retry

from kazu.utils.caching import kazu_disk_cache
from kazu.data.data import Mapping
from kazu.steps.linking.post_processing.mapping_strategies.strategies import MappingFactory

logger = logging.getLogger(__name__)


SourceOntology = str
SourceIdx = str

TargetSourceAndIdx = Tuple[str, str]
ToSourceAndIDXMap = Dict[SourceIdx, List[TargetSourceAndIdx]]
XrefDatabase = Dict[SourceOntology, ToSourceAndIDXMap]


def request_with_retry(url: str, headers: Dict, json_data: Dict):
    s = requests.Session()

    retries = Retry(
        total=25,
        backoff_factor=0.1,
        status_forcelist=[500, 502, 503, 504],
        allowed_methods=["POST"],
    )

    s.mount("https://", HTTPAdapter(max_retries=retries))

    return s.post(url=url, headers=headers, json=json_data)


class CrossReferenceManager(ABC):

    xref_db: XrefDatabase

    def __init__(self, source_to_parser_metadata_lookup: Dict[str, str], path: Path):
        """
        :param source_to_parser_metadata_lookup: when producing cross-referenced instances of Mapping, we need a
            reference in the MetadataDatabase to the target ontology, in order to look up the default label info etc.
            This lookup dict tells the cross reference manager what underlying parser it should use for a given source,
            since different parsers may hold sub sets or supersets of ids of each other. For example, a MedDRA hit
            might map to specific MONDO id. Since MONDO ids are held in both OpenTargetsDiseaseOntologyParser and
            MondoOntologyParser, we need to specify which one we want to use to generate the mapping
        :param path: path to cross ref mapping resources required by this manager
        """
        self.source_to_parser_metadata_lookup = source_to_parser_metadata_lookup
        self.xref_db = self.build_xref_cache(path)

    @abstractmethod
    def build_xref_cache(self, path: Path) -> XrefDatabase:
        """
        build a XrefDatabase suitable for caching

        :param path:
        :return:
        """
        pass

    def create_xref_mappings(self, mapping: Mapping) -> Iterable[Mapping]:
        """
        attempt to create additional xref mappings from a source mapping

        :param mapping:
        :return:
        """
        xref_lookup = self.xref_db.get(mapping.source, {})
        for (target_source, target_idx) in xref_lookup.get(mapping.idx, []):
            metadata_parser_name = self.source_to_parser_metadata_lookup.get(target_source)
            if metadata_parser_name is None:
                logger.warning(
                    f"source_to_parser_metadata_lookup not configured for target source: {metadata_parser_name}"
                )
                continue

            try:
                xref_mapping = MappingFactory.create_mapping(
                    parser_name=metadata_parser_name,
                    xref_source_parser_name=mapping.parser_name,
                    source=target_source,
                    idx=target_idx,
                    string_match_strategy=self.__class__.__name__,
                    disambiguation_strategy=mapping.disambiguation_strategy,
                    disambiguation_confidence=mapping.disambiguation_confidence,
                    string_match_confidence=mapping.string_match_confidence,
                    additional_metadata=mapping.metadata,
                )
                yield xref_mapping
            except KeyError:
                # note, we've set this to debug, as xref mapping sources (e.g. OXO) generally aren't version aware
                # i.e. this will probably fire a lot, and overload the logs if set anything higher than debug.
                # a custom log config at some point would probably help...
                logger.debug(
                    "failed to create xref mapping for %s->%s:%s->%s. Metadata not found.",
                    mapping.parser_name,
                    metadata_parser_name,
                    mapping.idx,
                    target_idx,
                )


class OxoCrossReferenceManager(CrossReferenceManager):
    """
    A CrossReferenceManager that uses the EBI OXO service to identify cross-references. Downloads a set of
    cross-references directly from EBI, and caches locally

    """

    oxo_url = "https://www.ebi.ac.uk/spot/oxo/api/search"
    headers = {"Content-Type": "application/json", "Accept": "application/json"}

    def __init__(
        self,
        source_to_parser_metadata_lookup: Dict[str, str],
        path: Path,
        oxo_kazu_name_mapping: Dict[str, str],
        uri_prefixes: Dict[str, str],
        oxo_query: Dict[str, List[str]],
    ):
        """

        :param source_to_parser_metadata_lookup:
        :param path:
        :param oxo_kazu_name_mapping: mapping of OXO source names to Kazu names, to covert OXO format to Kazu. If
            not specified, the OXO version will be used
        :param uri_prefixes: mapping of KAZU sources to URI prefixes, to correctly reconstruct ids. If not specified,
            no prefix will be used
        :param oxo_query: mapping of OXO source to target sources that will be used to construct the OXO API request
        """

        self.oxo_query = oxo_query
        self.uri_prefixes = uri_prefixes
        self.oxo_kazu_name_mapping = oxo_kazu_name_mapping
        super().__init__(source_to_parser_metadata_lookup, path)

    @kazu_disk_cache.memoize(ignore={0})
    def build_xref_cache(self, path: Path) -> XrefDatabase:
        oxo_dump_path = path.joinpath("oxo_dump.json")
        logger.info(f"looking for oxo dump at {oxo_dump_path}")
        if oxo_dump_path.exists():
            with open(oxo_dump_path, "r") as f:
                oxo_dump = json.load(f)
        else:
            logger.info(f"oxo dump not found. Attempting to download from {self.oxo_url}")
            oxo_dump = self.create_oxo_dump(oxo_dump_path)
        logger.info(f"loading from oxo dump at {oxo_dump_path}")
        xref_db = self.parse_oxo_dump(oxo_dump)
        return xref_db

    def _split_and_convert_curie(self, curie: str) -> Tuple[str, str]:
        """
        :param curie: a Compact uniform resource identifier, or CURIE. This is essentially a
            prefix followed by a colon (:) followed by a local ID
        :return: A source and idx, converted according to the class's oxo_kazu_name_mapping and uri_prefixes
        """
        oxo_source, oxo_idx = curie.split(":")
        converted_source = self.oxo_kazu_name_mapping.get(oxo_source, oxo_source)
        converted_idx = f"{self.uri_prefixes.get(converted_source,'')}{oxo_idx}"

        return converted_source, converted_idx

    def parse_oxo_dump(self, oxo_dump: List[Dict]) -> XrefDatabase:
        xref_db_default_dict: DefaultDict[
            str, DefaultDict[str, Set[Tuple[str, str]]]
        ] = defaultdict(lambda: defaultdict(set))

        for oxo_page in oxo_dump:
            for search_result in oxo_page["_embedded"]["searchResults"]:
                source, idx = self._split_and_convert_curie(search_result["curie"])
                xref_db_default_dict[source][idx].update(
                    self._split_and_convert_curie(mapping_response["curie"])
                    for mapping_response in search_result["mappingResponseList"]
                )
        xref_db = {}
        for source, id_to_xref in xref_db_default_dict.items():
            xref_db[source] = {idx: list(xref) for idx, xref in id_to_xref.items()}
        return xref_db

    def create_oxo_dump(self, path: Path) -> List[Dict]:
        results = []
        for input_source, mapping_target in self.oxo_query.items():
            data = {
                "ids": [],
                "inputSource": input_source,
                "mappingTarget": mapping_target,
                "distance": "1",
            }
            response = request_with_retry(
                url=f"{self.oxo_url}", headers=self.headers, json_data=data
            )
            response_data = response.json()
            link_info = response_data["_links"]
            last = link_info["last"]["href"]
            current = link_info["first"]["href"]

            while current != last:
                response = request_with_retry(url=current, headers=self.headers, json_data=data)
                response_data = response.json()
                link_info = response_data["_links"]
                next = link_info["next"]["href"]
                current = next
                results.append(response_data)
            else:
                response = request_with_retry(url=last, headers=self.headers, json_data=data)
                results.append(response.json())

        with open(path, "w") as f:
            json.dump(results, f)

        return results



import logging
import subprocess
import time
from typing import Any, Callable, Dict, List, Union, Optional

import hydra
import ray
from fastapi import Depends, FastAPI, HTTPException, Body
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from fastapi.security import HTTPBearer
from fastapi.staticfiles import StaticFiles
from hydra.utils import instantiate, call
from omegaconf import DictConfig
from pydantic import BaseModel
from ray import serve
from starlette.requests import HTTPConnection, Request
from starlette.responses import RedirectResponse

from kazu.data.data import Document, Entity
from kazu.pipeline import Pipeline, PipelineValueError
from kazu.utils.constants import HYDRA_VERSION_BASE
from kazu.web.routes import KAZU
from kazu.web.ls_web_utils import LSWebUtils


description = """
Welcome to the Web API of Kazu (Korea AstraZeneca University), a python biomedical NLP framework built in collaboration with Korea University,
designed to handle production workloads. This library aims to simplify the process of using state of the art NLP research in production systems. Some of the
research contained within are our own, but most of it comes from the community, for which we are immensely grateful.

The Web API is designed for light usage, if you need to run kazu for a heavy workload, please use the library directly. The Documentation for the library is available
*[here](https://astrazeneca.github.io/KAZU/index.html)*.
"""
logger = logging.getLogger("ray")
kazu_version = (
    subprocess.check_output("pip show kazu | grep Version", shell=True)
    .decode("utf-8")
    .split(" ")[1]
    .strip()
)
app = FastAPI(
    title="Kazu - Biomedical NLP Framework",
    description=description,
    version=kazu_version,
    license_info={
        "name": "Apache 2.0",
        "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
    },
)


def openapi_no_auth() -> Dict[str, Any]:
    """Remove the bits of the openapi schema that put auth buttons on the Swagger UI.

    When we don't configure any Authentication middleware, we otherwise still get the
    buttons, even though they aren't necessary to fill in and don't do anything, which
    may be confusing to users."""
    if not app.openapi_schema:
        base_schema = get_openapi(
            title=app.title,
            version=app.version,
            openapi_version=app.openapi_version,
            description=app.description,
            terms_of_service=app.terms_of_service,
            contact=app.contact,
            license_info=app.license_info,
            routes=app.routes,
            tags=app.openapi_tags,
            servers=app.servers,
        )
        del base_schema["components"]["securitySchemes"]
        for _path, path_data in base_schema["paths"].items():
            for _method, endpoint_data in path_data.items():
                # security in the openapi json is only present
                # to start with on those endpoints with
                # token=Depends(oauth2_scheme)
                if "security" in endpoint_data:
                    del endpoint_data["security"]
        app.openapi_schema = base_schema
    return app.openapi_schema


oauth2_scheme = HTTPBearer(auto_error=False)


def get_request_id(request: HTTPConnection) -> Union[str, None]:
    """Utility function for extracting custom header from HTTPConnection Object.

    :param request: Starlette HTTPConnection object
    :returns: ID string
    """
    key, req_id = request.headers.__dict__["_list"][-1]
    if key.decode("latin-1").lower() != "x-request-id":
        return None
    return req_id.decode("latin-1")


def get_id_log_prefix_if_available(request: HTTPConnection) -> str:
    """Utility function for generating the appropriate prefix for logs.

    :param request: Starlette HTTPConnection object
    :returns: Prefix to pre-pend to log messages containing the request id.
    """
    req_id = get_request_id(request)
    if req_id is not None:
        return "ID: " + req_id + " "
    else:
        return ""


def log_request_to_path_with_prefix(
    request: HTTPConnection, log_prefix: Optional[str] = None
) -> None:
    """Utility function to log the log prefix plus the endpoint the request was sent to.

    :param request: Starlette HTTPConnection object
    :param log_prefix: the prefix to log. Provided in case the log prefix has already been
        calculated in order to save re-calculation. Will call
        :func:`get_id_log_prefix_if_available` if not provided (or None is provided).
    """
    if log_prefix is None:
        log_prefix = get_id_log_prefix_if_available(request)
    logger.info(log_prefix + "Request to " + str(request.scope["path"]) + " endpoint")


_simple_doc_example = {
    "text": "A single string document that you want to recognise entities in."
    " Using the default kazu pipeline, this will recognise things like asthma, acetaminophen,"
    " EGFR and many others."
}
_sectioned_doc_example = {
    "sections": {
        "title": "a study about HER2 in breast cancer",
        "abstract": "We carried out a study on trastuzumab.",
        "fulltext": "A much longer text that probably mentions all of HER2,"
        " breast and gastric cancer, and trastuzumab.",
    }
}
_multiple_docs_mixed_type_example = [
    _simple_doc_example,
    _sectioned_doc_example,
    {"text": "Another simple doc, this one is about hypertension"},
    {
        "sections": {
            "first section": "A section about non-small cell lung cancer",
            "second section": "A section with nothing interesting in it",
            "final section": "A section about drugs: paracetamol, naproxin, ibuprofen.",
        }
    },
]


class SectionedWebDocument(BaseModel):
    sections: Dict[str, str]

    def to_kazu_document(self) -> Document:
        return Document.from_named_section_texts(self.sections)

    class Config:
        schema_extra = {"example": _sectioned_doc_example}


class SimpleWebDocument(BaseModel):
    text: str

    def to_kazu_document(self) -> Document:
        return Document.create_simple_document(self.text)

    class Config:
        schema_extra = {"example": _simple_doc_example}


WebDocument = Union[SimpleWebDocument, SectionedWebDocument]


class DocumentCollection(BaseModel):
    __root__: Union[List[WebDocument], WebDocument]

    def convert_to_kazu_documents(self) -> List[Document]:
        if isinstance(self.__root__, list):
            return [doc.to_kazu_document() for doc in self.__root__]
        else:
            return [self.__root__.to_kazu_document()]

    def __len__(self) -> int:
        if isinstance(self.__root__, list):
            return len(self.__root__)
        else:
            return 1

    class Config:
        schema_extra = {"example": _multiple_docs_mixed_type_example}


# You appear not to be able to provide multiple examples in the
# pydantic Config.schema_extra, so save this var for the Body
document_collection_examples = {
    "single_simple_doc": {
        "summary": "A single simple doc",
        "value": _simple_doc_example,
    },
    "multiple_simple_docs": {
        "summary": "Multiple simple docs",
        "value": [
            _simple_doc_example,
            {"text": "Another doc, this one is about hypertension"},
            {
                "text": "The last doc. This one doesn't have anything that should be picked up as an entity."
            },
        ],
    },
    "single_sectioned_doc": {"summary": "A single sectioned doc", "value": _sectioned_doc_example},
    "multiple_sectioned_docs": {
        "summary": "Multiple sectioned docs",
        "value": [
            _sectioned_doc_example,
            {
                "sections": {
                    "first section": "A section about non-small cell lung cancer",
                    "second section": "A section with nothing interesting in it",
                    "final section": "A section about drugs: paracetamol, naproxin, ibuprofen.",
                }
            },
        ],
    },
    "multiple_docs_mixed_types": {
        "summary": "Multiple docs, both simple and sectioned",
        "value": _multiple_docs_mixed_type_example,
    },
}
# the examples above don't make much sense for linking
linking_only_examples = {
    "single_simple_doc": {
        "summary": "A single simple doc",
        "value": {"text": "paracetamol"},
    },
    "multiple_simple_docs": {
        "summary": "Multiple simple docs",
        "value": [
            {"text": "paracetamol"},
            {"text": "acetaminophen"},
            {"text": "naproxin"},
        ],
    },
    "single_sectioned_doc": {
        "summary": "A single sectioned doc",
        "value": {"sections": {"first": "paracetamol", "second": "ibuprofen"}},
    },
    "multiple_sectioned_docs": {
        "summary": "Multiple sectioned docs",
        "value": [
            {"sections": {"first": "paracetamol", "second": "ibuprofen"}},
            {"sections": {"sec1": "naproxin", "sec2": "morphine"}},
        ],
    },
    "multiple_docs_mixed_types": {
        "summary": "Multiple docs, both simple and sectioned",
        "value": [
            {"text": "paracetamol"},
            {"sections": {"sec1": "naproxin", "sec2": "morphine"}},
            {"text": "acetaminophen"},
        ],
    },
}


class SingleEntityDocumentConverter:
    """Add an entity for the whole of every section in every document you provide.

    Essentially a subclass of :class:`DocumentCollection`, but pydantic makes this a pain to do with inheritance,
    so use composition instead of inheritance here."""

    def __init__(self, entity_free_doc_collection: DocumentCollection, entity_class: str) -> None:
        self.entity_class = entity_class
        self.entity_free_doc_collection = entity_free_doc_collection

    def convert_to_kazu_documents(self) -> List[Document]:
        documents = self.entity_free_doc_collection.convert_to_kazu_documents()
        for doc in documents:
            for section in doc.sections:
                section.entities = [
                    Entity.load_contiguous_entity(
                        # NOTE: our mapping logic is currently very dependent on the namespace
                        # I think though that we intend to get away from this with the 'mention_confidence' attribute
                        # on Entity and Curation
                        namespace="TransformersModelForTokenClassificationNerStep",
                        match=section.text,
                        entity_class=self.entity_class,
                        start=0,
                        end=len(section.text),
                    )
                ]
        return documents

    def __len__(self) -> int:
        return self.entity_free_doc_collection.__len__()


@serve.deployment(route_prefix="/api")
@serve.ingress(app)
class KazuWebAPI:
    """
    Web app to serve results
    """

    deploy: Callable

    def __init__(self, cfg: DictConfig):
        """
        :param cfg: DictConfig from Hydra
        """
        self.pipeline: Pipeline = instantiate(cfg.Pipeline)
        self.ls_web_utils: LSWebUtils = instantiate(cfg.LSWebUtils)
        if not cfg.Middlewares.auth_required:
            # override the openapi doc generating method
            # to remove authentication buttons.
            app.openapi = openapi_no_auth  # type: ignore[assignment]
            # we need the 'type: ignore' as otherwise mypy gives an error:
            # mypy doesn't like assigning to a method - it isn't able
            # to infer whether typing is consistent if you do this.
            # However, note that '@decorator' syntax is just syntactic
            # sugar for doing this, so this is a *fairly* reasonable
            # thing to do.

    @app.get("/")
    def get(self):
        logger.info("received request to root /")
        return "Welcome to KAZU."

    @app.get("/steps")
    def steps(self, request: Request):
        """Get a list of the steps in the the deployed pipeline."""
        log_request_to_path_with_prefix(request)
        return JSONResponse(content=list(self.pipeline._namespace_to_step))

    @app.get("/step_groups")
    def step_groups(self, request: Request):
        """Get the step groups configured in the deployed pipeline, (including showing the steps in each group)."""
        log_request_to_path_with_prefix(request)
        if self.pipeline.step_groups is None:
            return JSONResponse(content=None)
        return JSONResponse(
            content={
                group_name: [step.namespace() for step in steps]
                for group_name, steps in self.pipeline.step_groups.items()
            }
        )

    def base_pipeline_request(
        self,
        doc_collection: Union[DocumentCollection, SingleEntityDocumentConverter],
        request: Request,
        step_namespaces: Optional[List[str]] = None,
        step_group: Optional[str] = None,
    ):
        id_log_prefix = get_id_log_prefix_if_available(request)
        log_request_to_path_with_prefix(request, log_prefix=id_log_prefix)
        logger.info(id_log_prefix + "Documents sent: %s", len(doc_collection))
        try:
            result = self.pipeline(
                doc_collection.convert_to_kazu_documents(),
                step_namespaces=step_namespaces,
                step_group=step_group,
            )
        except PipelineValueError as e:
            raise HTTPException(status_code=422, detail=e.args[0]) from e
        return JSONResponse(content=[res.as_minified_dict() for res in result])

    @app.post(f"/{KAZU}/ner_and_linking")
    def ner_and_linking(
        self,
        request: Request,
        token=Depends(oauth2_scheme),
        doc_collection: DocumentCollection = Body(examples=document_collection_examples),
    ):
        """Run NER and Linking over the input document or documents.

        This is the default behaviour of the kazu library."""
        return self.base_pipeline_request(
            doc_collection=doc_collection, request=request, step_namespaces=None
        )

    @app.post(f"/{KAZU}/custom_pipeline_steps")
    def custom_pipeline_steps(
        self,
        request: Request,
        doc_collection: DocumentCollection,
        steps: Optional[List[str]] = None,
        token=Depends(oauth2_scheme),
    ):
        """Run specific steps over the provided document or documents.

        This is advanced functionality not expected to be needed by most ordinary users.

        The steps must be contained in the current pipeline (call the steps api to see what is
        available). The order the steps run in is the order provided in the API, not the order
        specified in the pipeline. Note that this means you can customize the order steps run in
        (although doing so effectively requires a good understanding of the steps and their
        behaviour)."""
        return self.base_pipeline_request(
            doc_collection=doc_collection,
            request=request,
            step_namespaces=steps,
        )

    @app.post(f"/{KAZU}/ner_only")
    def ner_only(
        self,
        request: Request,
        token=Depends(oauth2_scheme),
        doc_collection: DocumentCollection = Body(examples=document_collection_examples),
    ):
        """Call only steps that do Named Entity Recognition (NER).

        Note that this functionality is already available via the custom_steps endpoint,
        this just provides a convenient and visible/discoverable endpoint for users.
        """
        return self.base_pipeline_request(
            doc_collection=doc_collection, request=request, step_group="ner_only"
        )

    @app.post(f"/{KAZU}/linking_only")
    def linking_only(
        self,
        entity_class: str,
        request: Request,
        token=Depends(oauth2_scheme),
        doc_collection: DocumentCollection = Body(examples=linking_only_examples),
    ):
        """Call only steps that do Entity Linking (EL). Also known as 'entity normalization'.

        This API assumes that the whole of each section in each document is an entity mention. It
        will also try to link all entities to a single type of entity, specified in the
        entity_class query parameter. You can of course call the API multiple times with different
        entity_class values with the relevant documents for each entity_class.

        If you don't know what the entity type of the entities is, you can just call the
        ner_and_linking endpoint instead. It may however give you back entities that don't span the
        whole section.
        """
        linking_doc_collection = SingleEntityDocumentConverter(
            doc_collection, entity_class=entity_class
        )
        return self.base_pipeline_request(
            doc_collection=linking_doc_collection, request=request, step_group="linking_only"
        )

    # To be removed, after we check how often it gets called after being 'hidden'
    @app.post(f"/{KAZU}", deprecated=True)
    def ner(self, doc: WebDocument, request: Request, token=Depends(oauth2_scheme)):
        """Deprecated endpoint: use ner_and_linking.

        This endpoint will be removed in future.

        Behaves the same as ner_and_linking, but now we have the ner_only and
        linking_only endpoints, the naming here had the potential to be confusing."""
        id_log_prefix = get_id_log_prefix_if_available(request)
        logger.info(id_log_prefix + "Request to kazu endpoint")
        logger.info(id_log_prefix + "Document: %s", doc)
        result = self.pipeline([doc.to_kazu_document()])
        resp_dict = result[0].as_minified_dict()
        return JSONResponse(content=resp_dict)

    # To be removed, after we check how often it gets called after being 'hidden'
    @app.post(f"/{KAZU}/batch", deprecated=True)
    def batch_ner(self, docs: List[WebDocument], request: Request, token=Depends(oauth2_scheme)):
        """Deprecated endpoint: use ner_and_linking.

        This endpoint will be removed in future.

        Behaves the same as ner_and_linking, but now we have the ner_only and
        linking_only endpoints, the naming here had the potential to be confusing.

        In addition, we refactored ner_and_linking so that we no longer need separate
        'batch' and single document endpoints, now ner_and_linking handles both."""

        id_log_prefix = get_id_log_prefix_if_available(request)
        logger.info(id_log_prefix + "Request to kazu/batch endpoint")
        logger.info(id_log_prefix + "Documents sent: %s", len(docs))
        result = self.pipeline([doc.to_kazu_document() for doc in docs])
        return JSONResponse(content=[res.as_minified_dict() for res in result])

    @app.post(f"/{KAZU}/ls-annotations")
    def ls_annotations(self, doc: WebDocument, request: Request):
        """Provide LabelStudio annotations from the kazu results on the given document.

        This is not expected to be called by ordinary users. This endpoint was added in order
        to support the Kazu frontend, which makes uses of LabelStudio's frontend code, so
        relies on this endpoint."""

        id_log_prefix = get_id_log_prefix_if_available(request)
        log_request_to_path_with_prefix(request, log_prefix=id_log_prefix)
        logger.info(id_log_prefix + "Document: %s", doc)
        result = self.pipeline([doc.to_kazu_document()])[0]
        ls_view, ls_tasks = self.ls_web_utils.kazu_doc_to_ls(result)
        return JSONResponse(
            content={"ls_view": ls_view, "ls_tasks": ls_tasks, "doc": result.as_minified_dict()}
        )

    # Note: this needs to be defined last so that we don't try and
    # interpret the ls-annotations or batch API calls as step groups.
    @app.post(f"/{KAZU}/{{step_group}}")
    def step_group(
        self,
        step_group: str,
        request: Request,
        token=Depends(oauth2_scheme),
        doc_collection: DocumentCollection = Body(examples=document_collection_examples),
    ):
        """Run the pipeline with a specific step group.

        This will run only a pre-defined subset of steps. Call the step_groups endpoint to
        find the configured set of step_groups for this API deployment, and the steps they
        include.

        If you are interested in running a different set of steps, you can use the
        custom_pipeline_steps endpoint.
        """
        return self.base_pipeline_request(
            doc_collection=doc_collection,
            request=request,
            step_group=step_group,
        )


@serve.deployment(route_prefix="/ui")
@serve.ingress(app)
class KazuWebUI:

    deploy: Callable

    def __init__(self, ui_cfg: DictConfig):
        app.mount(
            ui_cfg.baseURL,
            StaticFiles(directory=ui_cfg.staticFilesPath, html=True),
            name="static-ui-content",
        )

    @app.get("/")
    def root(self):
        return RedirectResponse("/index.html")


@hydra.main(version_base=HYDRA_VERSION_BASE, config_path="../conf", config_name="config")
def start(cfg: DictConfig) -> None:
    """
    deploy the web app to Ray Serve

    :param cfg: DictConfig from Hydra
    :return: None
    """
    # Connect to the running Ray cluster, or run as single node
    call(cfg.ray.init)
    call(cfg.ray.serve)

    KazuWebAPI.deploy(cfg)
    if (ui_conf := cfg.ray.get("ui")) is not None:
        KazuWebUI.deploy(ui_conf)

    if not cfg.ray.serve.detached:
        while True:
            logger.info(serve.list_deployments())
            time.sleep(600)


def stop():
    if ray.is_initialized():
        serve.shutdown()
        ray.shutdown()


if __name__ == "__main__":
    start()



import logging
from collections import defaultdict
from typing import List, Set, Optional

from kazu.data.data import Document, SynonymTermWithMetrics
from kazu.steps.step import Step, document_batch_step
from kazu.utils.caching import EntityLinkingLookupCache
from kazu.utils.grouping import sort_then_group
from kazu.utils.link_index import DictionaryIndex

logger = logging.getLogger(__name__)


class DictionaryEntityLinkingStep(Step):
    """Uses :class:`kazu.utils.link_index.DictionaryIndex` to match entities to ontologies.

    Note, this is not an instance of :class:`kazu.steps.step.ParserDependentStep`, as
    this logic would duplicate the work of :class:`kazu.utils.link_index.DictionaryIndex`
    """

    def __init__(
        self,
        indices: List[DictionaryIndex],
        lookup_cache_size: int = 5000,
        top_n: int = 20,
        skip_ner_namespaces: Optional[Set[str]] = None,
    ):
        """

        :param indices: indices to query
        :param lookup_cache_size: the size of the Least Recently Used lookup cache to maintain
        :param top_n: keep the top_n results for the query (passed to :class:`kazu.utils.link_index.DictionaryIndex`)
        :param skip_ner_namespaces: set of NER-step namespaces -- linking will be skipped for entities generated by
            these namespaces
        """
        self.entity_class_to_indices = defaultdict(set)
        for index in indices:
            self.entity_class_to_indices[index.entity_class].add(index)
        self.top_n = top_n
        self.skip_ner_namespaces = skip_ner_namespaces if skip_ner_namespaces is not None else set()
        self.lookup_cache = EntityLinkingLookupCache(lookup_cache_size)

    @document_batch_step
    def __call__(self, docs: List[Document]):
        """
        logic of entity linker:

        1. first obtain an entity list from all docs
        2. check the lookup LRUCache to see if an entity has been recently processed
        3. if the cache misses, run a string similarity search using the configured :class:`kazu.utils.link_index.DictionaryIndex` 's

        :param docs:
        :return:
        """
        entities = (
            ent
            for doc in docs
            for ent in doc.get_entities()
            if ent.namespace not in self.skip_ner_namespaces
        )
        ents_by_match_and_class = {
            k: list(v) for k, v in sort_then_group(entities, lambda x: (x.match, x.entity_class))
        }
        if len(ents_by_match_and_class) > 0:
            for ent_match_and_class, ents_this_match in ents_by_match_and_class.items():
                cache_missed_entities = self.lookup_cache.check_lookup_cache(ents_this_match)
                if len(cache_missed_entities) == 0:
                    continue
                else:
                    indices_to_search = self.entity_class_to_indices.get(ent_match_and_class[1])
                    if indices_to_search:
                        terms: List[SynonymTermWithMetrics] = []
                        for index in indices_to_search:
                            terms.extend(index.search(ent_match_and_class[0], self.top_n))

                        for ent in ents_this_match:
                            ent.update_terms(terms)

                        self.lookup_cache.update_terms_lookup_cache(
                            entity=next(iter(ents_this_match)), terms=terms
                        )



import logging
from copy import deepcopy
from typing import Optional, Dict, List, Tuple, Set, Iterable

from kazu.data.data import (
    SynonymTerm,
    SimpleValue,
    EquivalentIdAggregationStrategy,
    AssociatedIdSets,
)
from kazu.utils.utils import Singleton

logger = logging.getLogger(__name__)

# type aliases to make function signatures more explanatory,
# reduce need for comments.
ParserName = str
Idx = str
NormalisedSynonymStr = str
Metadata = Dict[str, SimpleValue]


class MetadataDatabase(metaclass=Singleton):
    """
    Singleton of Ontology metadata database. Purpose: metadata needs to be looked up in different linking processes,
    and this singleton allows us to load it once/reduce memory usage
    """

    def __init__(self):
        self._database: Dict[ParserName, Dict[Idx, Metadata]] = {}
        self._keys_lst: Dict[ParserName, List[Idx]] = {}
        self.loaded_parsers: Set[str] = set()

    def add_parser(self, name: ParserName, metadata: Dict[Idx, Metadata]):
        """
        add metadata to the ontology. Note, metadata is assumed to be static, and global. Calling this function will
        override any existing entries with associated with the keys in the metadata dict

        :param name: name of ontology to add to
        :param metadata: dict in format {idx:metadata}
        :return:
        """
        self.loaded_parsers.add(name)
        if name in self._database:
            logger.info(
                f"parser {name} already present in metadata database - will override existing parser data."
            )
        self._database[name] = metadata
        self._keys_lst[name] = list(self._database[name].keys())

    def get_by_idx(self, name: ParserName, idx: Idx) -> Metadata:
        """
        get the metadata associated with an ontology and id

        :param name: name of ontology to query
        :param idx: idx to query
        :return:
        """
        return deepcopy(self._database[name][idx])

    def get_by_index(self, name: ParserName, i: int) -> Tuple[Idx, Metadata]:
        idx = self._keys_lst[name][i]
        return deepcopy(
            (
                idx,
                self._database[name][idx],
            )
        )

    def get_all(self, name: ParserName) -> Dict[Idx, Metadata]:
        """
        get all metadata associated with an ontology

        :param name: name of ontology
        :return:
        """
        return self._database[name]


class SynonymDatabase(metaclass=Singleton):
    """
    Singleton of a database of synonyms.
    """

    def __init__(self):
        self._syns_database_by_syn: Dict[ParserName, Dict[NormalisedSynonymStr, SynonymTerm]] = {}
        self._syns_by_aggregation_strategy: Dict[
            ParserName, Dict[EquivalentIdAggregationStrategy, Dict[Idx, Set[NormalisedSynonymStr]]]
        ] = {}
        self._associated_id_sets_by_id: Dict[ParserName, Dict[str, Set[AssociatedIdSets]]] = {}
        self.loaded_parsers: Set[ParserName] = set()

    def add(self, name: ParserName, synonyms: Iterable[SynonymTerm]) -> None:
        """
        add synonyms to the database.

        :param name: name of ontology to add to
        :param synonyms: iterable of SynonymTerms to add
        :return:
        """
        self.loaded_parsers.add(name)
        if name not in self._syns_database_by_syn:
            self._syns_database_by_syn[name] = {}
            self._associated_id_sets_by_id[name] = {}
        for synonym in synonyms:
            self._syns_database_by_syn[name][synonym.term_norm] = synonym
            for equiv_ids in synonym.associated_id_sets:
                for idx in equiv_ids.ids:
                    dict_for_this_parser = self._syns_by_aggregation_strategy.setdefault(name, {})
                    dict_for_this_aggregation_strategy = dict_for_this_parser.setdefault(
                        synonym.aggregated_by, {}
                    )
                    syn_set_for_this_id = dict_for_this_aggregation_strategy.setdefault(idx, set())
                    syn_set_for_this_id.add(synonym.term_norm)
                    self._associated_id_sets_by_id[name].setdefault(idx, set()).add(
                        synonym.associated_id_sets
                    )

    def get(self, name: ParserName, synonym: NormalisedSynonymStr) -> SynonymTerm:
        """
        get a set of EquivalentIdSets associated with an ontology and synonym string

        :param name: name of ontology to query
        :param synonym: idx to query
        :return:
        """
        return self._syns_database_by_syn[name][synonym]

    def get_syns_for_id(
        self,
        name: ParserName,
        idx: Idx,
        strategy_filters: Optional[Set[EquivalentIdAggregationStrategy]] = None,
    ) -> Set[NormalisedSynonymStr]:
        result = set()
        if strategy_filters is None:
            for syn_dict in self._syns_by_aggregation_strategy[name].values():
                result.update(syn_dict.get(idx, set()))
        else:
            for agg_strategy in strategy_filters:
                result.update(
                    self._syns_by_aggregation_strategy[name]
                    .get(agg_strategy, dict())
                    .get(idx, set())
                )
        return result

    def get_syns_sharing_id(
        self,
        name: ParserName,
        synonym: NormalisedSynonymStr,
        strategy_filters: Optional[Set[EquivalentIdAggregationStrategy]] = None,
    ) -> Set[NormalisedSynonymStr]:
        """
        get all other syns for a synonym in a kb

        :param name: parser name
        :param synonym: synonym
        :param strategy_filters: Optional set of EquivalentIdAggregationStrategy. If provided, only syns aggregated
            via these strategies will be returned. If None (the default), all syns will be returned
        :return:
        """
        result: Set[str] = set()
        synonym_term = self.get(name, synonym)
        if strategy_filters is not None and synonym_term.aggregated_by not in strategy_filters:
            return result
        for equiv_id_set in synonym_term.associated_id_sets:
            for idx in equiv_id_set.ids:
                syns = self.get_syns_for_id(
                    name,
                    idx,
                    strategy_filters,
                )
                result.update(syns)
        return result

    def get_all(self, name: ParserName) -> Dict[NormalisedSynonymStr, SynonymTerm]:
        """
        get all synonyms associated with an ontology

        :param name: name of ontology
        :return:
        """
        return self._syns_database_by_syn[name]



import dataclasses
import json
import uuid
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime, date
from enum import Enum, auto, IntEnum
from math import inf
from typing import List, Any, Dict, Optional, Tuple, FrozenSet, Set, Iterable, Union, DefaultDict

import bson
from bson import json_util
from numpy import ndarray, float32, float16

from kazu.utils.string_normalizer import StringNormalizer

IS_SUBSPAN = "is_subspan"
# BIO schema
ENTITY_START_SYMBOL = "B"
ENTITY_INSIDE_SYMBOL = "I"
ENTITY_OUTSIDE_SYMBOL = "O"

# key for Document Processing Failed
PROCESSING_EXCEPTION = "PROCESSING_EXCEPTION"


NoneType = type(None)
JsonDictType = Union[Dict[str, Any], List, int, float, bool, str, NoneType]


class AutoNameEnum(Enum):
    """Subclass to create an Enum where values are the names when using :class:`enum.auto`\\ .

    Taken from the `Python Enum Docs <https://docs.python.org/3/howto/enum.html#using-automatic-values>`_.

    This is `licensed under Zero-Clause BSD. <https://docs.python.org/3/license.html#zero-clause-bsd-license-for-code-in-the-python-release-documentation>`_

    .. raw:: html

        <details>
        <summary>Full License</summary>

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.

    .. raw:: html

        </details>
    """

    def _generate_next_value_(name, start, count, last_values):
        return name


class MentionConfidence(IntEnum):
    HIGHLY_LIKELY = 100  # almost certain to be correct
    PROBABLE = 50
    POSSIBLE = 10  # high degree of uncertainty


class StringMatchConfidence(AutoNameEnum):
    HIGHLY_LIKELY = auto()  # almost certain to be correct
    PROBABLE = auto()  # on the balance of probabilities, will be correct
    POSSIBLE = auto()  # high degree of uncertainty


class DisambiguationConfidence(AutoNameEnum):
    HIGHLY_LIKELY = auto()  # almost certain to be correct
    PROBABLE = auto()  # on the balance of probabilities, will be correct
    POSSIBLE = auto()  # high degree of uncertainty
    AMBIGUOUS = auto()  # could not disambiguate


@dataclass(frozen=True)
class CharSpan:
    """A concept similar to a Spacy Span, except is character index based rather than token based"""

    start: int
    end: int

    def is_completely_overlapped(self, other):
        """
        True if other completely overlaps this span

        :param other:
        :return:
        """
        return self.start >= other.start and self.end <= other.end

    def is_partially_overlapped(self, other):
        """
        True if other partially overlaps this span

        :param other:
        :return:
        """
        return (other.start <= self.start <= other.end) or (other.start <= self.end <= other.end)

    def __lt__(self, other):
        return self.start < other.start

    def __gt__(self, other):
        return self.end > other.end


class EquivalentIdAggregationStrategy(AutoNameEnum):
    NO_STRATEGY = auto()  # no strategy. should be used for debugging/testing only
    RESOLVED_BY_SIMILARITY = auto()  # synonym linked to ID via similarity to default ID label
    SYNONYM_IS_AMBIGUOUS = auto()  # synonym has no unambiguous meaning
    CUSTOM = auto()  # a place holder for any strategy that
    UNAMBIGUOUS = auto()
    MERGED_AS_NON_SYMBOLIC = auto()  # used when non-symbolic synonyms are merged
    MODIFIED_BY_CURATION = auto()


@dataclass(frozen=True, eq=True, order=True)
class EquivalentIdSet:
    """
    A representation of a set of kb ID's that map to the same synonym and mean the same thing.
    """

    ids_and_source: FrozenSet[Tuple[str, str]] = field(
        default_factory=frozenset, hash=True
    )  # other ID's mapping to this syn, from different KBs

    @property
    def sources(self) -> Set[str]:
        return set(x[1] for x in self.ids_and_source)

    @property
    def ids(self) -> Set[str]:
        return set(x[0] for x in self.ids_and_source)


@dataclass(frozen=True)
class Mapping:
    """
    a mapping is a fully mapped and disambiguated kb concept
    """

    #: default label from knowledgebase
    default_label: str
    #: the knowledgebase/database/ontology name
    source: str
    #: the origin of this mapping
    parser_name: str
    #: the identifier within the KB
    idx: str
    string_match_strategy: str
    string_match_confidence: StringMatchConfidence
    disambiguation_confidence: Optional[DisambiguationConfidence] = None
    disambiguation_strategy: Optional[str] = None
    #: source parser name if mapping is an XREF
    xref_source_parser_name: Optional[str] = None
    #: generic metadata
    metadata: Dict[Any, Any] = field(default_factory=dict, hash=False)


NumericMetric = Union[bool, int, float]

AssociatedIdSets = FrozenSet[EquivalentIdSet]
"""A frozen set of :class:`.EquivalentIdSet`"""


@dataclass(frozen=True, eq=True)
class SynonymTerm:
    """
    a SynonymTerm is a container for a single normalised synonym, and is produced by an OntologyParser implementation.
    It may be composed of multiple terms that normalise
    to the same unique string (e.g. "breast cancer" and "Breast Cancer"). The number of associated_id_sets that this
    synonym maps to is determined by the  score_and_group_ids method of the associated OntologyParser
    """

    #: unnormalised synonym strings
    terms: FrozenSet[str]
    #: normalised form
    term_norm: str
    #: ontology parser name
    parser_name: str
    #: is the term symbolic? Determined by the OntologyParser
    is_symbolic: bool
    associated_id_sets: AssociatedIdSets
    #: mapping type metadata
    mapping_types: FrozenSet[str] = field(hash=False)
    #: aggregation strategy, determined by the ontology parser
    aggregated_by: EquivalentIdAggregationStrategy = field(hash=False, compare=False)
    #: if this term was generated by a synonym generator, this is the original source synonym
    original_term: Optional[str] = None

    @property
    def is_ambiguous(self):
        return len(self.associated_id_sets) > 1


@dataclass(frozen=True, eq=True)
class SynonymTermWithMetrics(SynonymTerm):
    """
    Similar to SynonymTerm, but also allows metrics to be scored. As these metrics are not used in the hash function,
    care should be taken when hashing of this object is required
    """

    search_score: Optional[float] = field(compare=False, default=None)
    embed_score: Optional[float] = field(compare=False, default=None)
    bool_score: Optional[float] = field(compare=False, default=None)
    exact_match: Optional[bool] = field(compare=False, default=None)

    @staticmethod
    def from_synonym_term(
        term: SynonymTerm,
        search_score: Optional[float] = None,
        embed_score: Optional[float] = None,
        bool_score: Optional[float] = None,
        exact_match: Optional[bool] = None,
    ):

        return SynonymTermWithMetrics(
            search_score=search_score,
            embed_score=embed_score,
            bool_score=bool_score,
            exact_match=exact_match,
            **term.__dict__,
        )

    def merge_metrics(self, term: "SynonymTermWithMetrics") -> "SynonymTermWithMetrics":
        new_values = {
            k: v
            for k, v in term.__dict__.items()
            if k in {"search_score", "embed_score", "bool_score", "exact_match"} and v is not None
        }
        new_term = dataclasses.replace(self, **new_values)
        return new_term


@dataclass
class Entity:
    """
    A :class:`kazu.data.data.Entity` is a container for information about a single entity detected within a :class:`kazu.data.data.Section`

    Within an :class:`kazu.data.data.Entity`, the most important fields are :attr:`.Entity.match` (the actual string detected),
    :attr:`.Entity.syn_term_to_synonym_terms`, a dict of :class:`kazu.data.data.SynonymTermWithMetrics` (candidates for knowledgebase hits)
    and :attr:`.Entity.mappings`, the final product of linked references to the underlying entity

    """

    #: exact text representation
    match: str
    entity_class: str
    spans: FrozenSet[CharSpan]
    #: # namespace of the :class:`~.Step` that produced this instance
    namespace: str
    mention_confidence: MentionConfidence = MentionConfidence.HIGHLY_LIKELY
    mappings: Set[Mapping] = field(default_factory=set)
    #: generic metadata
    metadata: Dict[Any, Any] = field(default_factory=dict)
    start: int = field(init=False)
    end: int = field(init=False)
    match_norm: str = field(init=False)
    syn_term_to_synonym_terms: Dict[SynonymTermWithMetrics, SynonymTermWithMetrics] = field(
        default_factory=dict
    )

    def update_terms(self, terms: Iterable[SynonymTermWithMetrics]):
        for term in terms:
            existing_term: Optional[SynonymTermWithMetrics] = self.syn_term_to_synonym_terms.get(
                term
            )
            if existing_term is not None:
                new_term = existing_term.merge_metrics(term)
                self.syn_term_to_synonym_terms[new_term] = new_term
            else:
                self.syn_term_to_synonym_terms[term] = term

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return id(self) == id(other)

    def calc_starts_and_ends(self) -> Tuple[int, int]:
        earliest_start = inf
        latest_end = 0
        for span in self.spans:
            if span.start < earliest_start:
                earliest_start = span.start
            if span.end > latest_end:
                latest_end = span.end
        if earliest_start is inf:
            raise ValueError("spans are not valid")
        return int(earliest_start), latest_end

    def __post_init__(self):
        self.start, self.end = self.calc_starts_and_ends()
        self.match_norm = StringNormalizer.normalize(self.match, self.entity_class)

    def is_completely_overlapped(self, other):
        """
        True if all CharSpan instances are completely encompassed by all other CharSpan instances

        :param other:
        :return:
        """
        for charspan in self.spans:
            for other_charspan in other.spans:
                if charspan.is_completely_overlapped(other_charspan):
                    break
            else:
                return False
        return True

    def is_partially_overlapped(self, other):
        """
        True if only one CharSpan instance is defined in both self and other, and they are partially overlapped

        If multiple CharSpan are defined in both self and other, this becomes pathological, as while they may overlap
        in the technical sense, they may have distinct semantic meaning. For instance, consider the case where
        we may want to use is_partially_overlapped to select the longest annotation span suggested
        by some NER system.

        case 1:
        text: the patient has metastatic liver cancers
        entity1:  metastatic liver cancer -> [CharSpan(16,39]
        entity2: liver cancers -> [CharSpan(27,40]

        result: is_partially_overlapped -> True (entities are part of same concept)


        case 2: non-contiguous entities

        text: lung and liver cancer
        lung cancer -> [CharSpan(0,4), CharSpan(1521
        liver cancer -> [CharSpan(9,21)]

        result: is_partially_overlapped -> False (entities are distinct)

        :param other:
        :return:
        """
        if len(self.spans) == 1 and len(other.spans) == 1:
            (charspan,) = self.spans
            (other_charspan,) = other.spans
            return charspan.is_partially_overlapped(other_charspan)
        else:
            return False

    def __len__(self) -> int:
        """
        Span length

        :return: number of characters enclosed by span
        """
        return self.end - self.start

    def __repr__(self) -> str:
        """
        Describe the tag

        :return: tag match description
        """
        return f"{self.match}:{self.entity_class}:{self.namespace}:{self.start}:{self.end}"

    def as_brat(self):
        """
        :return: self as the third party biomedical nlp Brat format, (see docs on Brat)
        """
        # TODO: update this to make use of non-contiguous entities
        return f"{hash(self)}\t{self.entity_class}\t{self.start}\t{self.end}\t{self.match}\n"

    def add_mapping(self, mapping: Mapping):
        """
        deprecated

        :param mapping:
        :return:
        """
        self.mappings.add(mapping)

    @classmethod
    def from_spans(cls, spans: List[Tuple[int, int]], text: str, join_str: str = "", **kwargs):
        """
        create an instance of Entity from a list of character indices. A text string of underlying doc is
        also required to produce a representative match

        :param spans:
        :param text:
        :param join_str: a string used to join the spans together
        :param kwargs:
        :return:
        """
        char_spans = []
        text_pieces = []
        for start, end in spans:
            text_pieces.append(text[start:end])
            char_spans.append(CharSpan(start=start, end=end))
        return cls(spans=frozenset(char_spans), match=join_str.join(text_pieces), **kwargs)

    @classmethod
    def load_contiguous_entity(cls, start: int, end: int, **kwargs) -> "Entity":
        single_span = frozenset([CharSpan(start=start, end=end)])
        return cls(spans=single_span, **kwargs)


@dataclass(unsafe_hash=True)
class Section:
    #: the text to be processed
    text: str
    #: the name of the section (e.g. abstract, body, header, footer etc)
    name: str
    #: not required. if a preprocessed_text is used, this represents mappings of the preprocessed charspans back to the original
    offset_map: Dict[CharSpan, CharSpan] = field(default_factory=dict, hash=False, init=False)
    #: generic metadata
    metadata: Dict[Any, Any] = field(default_factory=dict, hash=False)
    #: entities detected in this section
    entities: List[Entity] = field(default_factory=list, hash=False)
    _sentence_spans: Optional[Tuple[CharSpan, ...]] = field(
        default=None, hash=False, init=False
    )  # hidden implem. to prevent overwriting existing sentence spans

    @property
    def sentence_spans(self) -> Iterable[CharSpan]:
        if self._sentence_spans is not None:
            return self._sentence_spans
        else:
            return ()

    @sentence_spans.setter
    def sentence_spans(self, sent_spans: Iterable[CharSpan]):
        """
        Setter for sentence_spans. sentence_spans are stored in the order provided by the iterable
        sent_spans param, which may not necessarily be in sorted order.

        :param sent_spans:
        :return:
        """
        if self._sentence_spans is None:
            sent_spans_tuple = tuple(sent_spans)
            assert len(sent_spans_tuple) == len(
                set(sent_spans_tuple)
            ), "There are duplicate sentence spans"
            self._sentence_spans = sent_spans_tuple
        else:
            raise AttributeError("Immutable sentence_spans is already set")

    def __str__(self):
        return f"name: {self.name}, text: {self.text[:100]}"


@dataclass(unsafe_hash=True)
class Document:
    #: a document identifier
    idx: str = field(default_factory=lambda: uuid.uuid4().hex)
    #: sections comprising this document
    sections: List[Section] = field(default_factory=list, hash=False)
    #: generic metadata
    metadata: Dict[Any, Any] = field(default_factory=dict, hash=False)

    def __str__(self):
        return f"idx: {self.idx}"

    def get_entities(self) -> List[Entity]:
        """
        get all entities in this document
        """
        entities = []
        for section in self.sections:
            entities.extend(section.entities)
        return entities

    def json(
        self,
        drop_unmapped_ents: bool = False,
        drop_terms: bool = False,
        **kwargs,
    ):
        """
        custom encoder needed to handle serialisation issues with our data model

        :param: drop_unmapped_ents: drop any entities that have no mappings
        :param: drop_terms: drop the synonym term dict field
        :param kwargs: additional kwargs passed to json.dumps
        :return:
        """
        as_dict = self.as_minified_dict(
            drop_unmapped_ents=drop_unmapped_ents, drop_terms=drop_terms
        )
        return json.dumps(as_dict, **kwargs)

    def as_minified_dict(self, drop_unmapped_ents: bool = False, drop_terms: bool = False) -> Dict:
        as_dict = DocumentJsonUtils.doc_to_json_dict(self)
        as_dict = DocumentJsonUtils.minify_json_dict(
            as_dict, drop_unmapped_ents=drop_unmapped_ents, drop_terms=drop_terms
        )
        return DocumentJsonUtils.remove_empty_elements(as_dict)

    @classmethod
    def create_simple_document(cls, text: str) -> "Document":
        """
        Create an instance of :class:`.Document` from a text string.

        :param text:
        :return:
        """
        return cls(sections=[Section(text=text, name="na")])

    @classmethod
    def simple_document_from_sents(cls, sents: List[str]) -> "Document":
        section = Section(text=" ".join(sents), name="na")
        sent_spans = []
        curr_start = 0
        for sent in sents:
            sent_spans.append(CharSpan(start=curr_start, end=curr_start + len(sent)))
            curr_start += len(sent) + 1  # + 1 is for the joining space
        section.sentence_spans = sent_spans
        return cls(sections=[section])

    @classmethod
    def from_named_section_texts(cls, named_sections: Dict[str, str]) -> "Document":
        sections = [Section(text=text, name=name) for name, text in named_sections.items()]
        return cls(sections=sections)

    def __len__(self):
        length = 0
        for section in self.sections:
            length += len(section.text)
        return length


class DocumentJsonUtils:
    class ConversionException(Exception):
        pass

    atomic_types = (int, float, str, bool, type(None))
    listlike_types = (list, tuple, set, frozenset)

    @staticmethod
    def minify_json_dict(
        doc_json_dict: Dict[str, Any],
        drop_unmapped_ents: bool = False,
        drop_terms: bool = False,
        in_place=True,
    ) -> Dict:
        doc_json_dict = doc_json_dict if in_place else deepcopy(doc_json_dict)

        if drop_unmapped_ents or drop_terms:
            for section_dict in doc_json_dict["sections"]:
                section_entities = section_dict["entities"]
                ents_to_keep = list(
                    filter(lambda _ent: _ent["mappings"], section_entities)
                    if drop_unmapped_ents
                    else section_entities
                )
                if drop_terms:
                    for ent in ents_to_keep:
                        ent["synonym_terms"].clear()

                section_dict["entities"] = ents_to_keep

        return doc_json_dict

    @classmethod
    def doc_to_json_dict(cls, doc: Document) -> Dict[str, JsonDictType]:
        return {k: DocumentJsonUtils.obj_to_dict_repr(v) for k, v in doc.__dict__.items()}

    @classmethod
    def obj_to_dict_repr(cls, obj: Any) -> JsonDictType:
        if isinstance(obj, Enum):
            return obj.name
        elif isinstance(obj, cls.atomic_types):
            return obj
        elif isinstance(obj, (float16, float32)):  # type: ignore[misc]
            return obj.item()
        elif isinstance(obj, cls.listlike_types):
            return [cls.obj_to_dict_repr(elem) for elem in obj]
        elif isinstance(obj, ndarray):
            return obj.tolist()
        elif dataclasses.is_dataclass(obj):
            return cls.obj_to_dict_repr(obj.__dict__)
        elif isinstance(obj, dict):
            processed_dict_pairs = (cls._preprocess_dict_pair(pair) for pair in obj.items())
            return {k: cls.obj_to_dict_repr(v) for k, v in processed_dict_pairs}
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif isinstance(obj, bson.ObjectId):
            return json_util.default(obj)
        else:
            raise cls.ConversionException(f"Unknown object type: {type(obj)}")

    @classmethod
    def _preprocess_dict_pair(cls, kv_pair: Tuple[Any, Any]) -> Tuple[str, JsonDictType]:
        k, v = kv_pair
        if k == "offset_map":
            # v is a dict of CharSpan->CharSpan, needs conversion
            return k, list(v.items())
        elif k == "_sentence_spans":
            return "sentence_spans", list(v)
        elif k == "syn_term_to_synonym_terms":
            return "synonym_terms", list(v.values())
        else:
            return k, v

    @staticmethod
    def remove_empty_elements(d: Any):
        """recursively remove empty lists, empty dicts, or None elements from a dictionary"""
        if not isinstance(d, (dict, list)):
            return d
        elif isinstance(d, list):
            return [
                v
                for v in (DocumentJsonUtils.remove_empty_elements(v) for v in d)
                if not DocumentJsonUtils.empty(v)
            ]
        else:
            return {
                k: v
                for k, v in ((k, DocumentJsonUtils.remove_empty_elements(v)) for k, v in d.items())
                if not DocumentJsonUtils.empty(v)
            }

    @staticmethod
    def empty(x) -> bool:
        return x is None or x == {} or x == []


SimpleValue = Union[NumericMetric, str]


class CuratedTermBehaviour(AutoNameEnum):
    #: do nothing. Useful if you want to catalogue a term as "I've looked at this but don't want to use it for NER"
    IGNORE = auto()
    #: use the term for both dictionary based NER and as a linking target.
    ADD_FOR_NER_AND_LINKING = auto()
    #: use the term only as a linking target. Note, this is not required if the term is already in the
    #: underlying ontology, as all ontology terms are included as linking targets by default
    #: (Also see DROP_SYNONYM_TERM_FOR_LINKING)
    ADD_FOR_LINKING_ONLY = auto()
    #: do not use this term as a linking target. Normally, you would use this for a term you want to remove
    #: from the underlying ontology (e.g. a 'bad' synonym). If the term does not exist, has no effect
    DROP_SYNONYM_TERM_FOR_LINKING = auto()
    #: used if the containing Curation has source_term !=None, in which case all behaviours are inherited
    INHERIT_FROM_SOURCE_TERM = auto()


class ParserBehaviour(AutoNameEnum):
    #: completely remove the ids from the parser - i.e. should never used anywhere
    DROP_IDS_FROM_PARSER = auto()


@dataclass(frozen=True)
class ParserAction:
    """
    A ParserAction changes the behaviour of a :class:`kazu.ontology_preprocessing.base.OntologyParser` in
    a global sense.

    A ParserAction overrides any default behaviour of the parser, and also any conflicts that may occur with
    :class:`.CuratedTerm`\\s.

    These actions are useful for eliminating unwanted behaviour. For example, the root of the Mondo
    ontology is http://purl.obolibrary.org/obo/HP_0000001, which has a default label of 'All'. Since this is
    such a common word, and not very useful in terms of linking, we might want a global action so that this
    ID is not used anywhere in a Kazu pipeline.

    The parser_to_target_id_mappings field should specify the parser name and an affected IDs if required.
    See :class:`.ParserBehaviour` for the type of actions that are possible.

    """

    behaviour: ParserBehaviour
    parser_to_target_id_mappings: Dict[str, Set[str]] = field(default_factory=dict)

    @classmethod
    def from_json(cls, json_dict: Dict) -> "ParserAction":
        return cls(
            behaviour=ParserBehaviour(json_dict["behaviour"]),
            parser_to_target_id_mappings={
                k: set(v) for k, v in json_dict["parser_to_target_id_mappings"].items()
            },
        )

    def __post_init__(self):
        if len(self.parser_to_target_id_mappings) == 0:
            raise ValueError(
                f"parser_to_target_id_mappings must be specified for global actions: {self}"
            )
        for key, values in self.parser_to_target_id_mappings.items():
            if len(values) == 0:
                raise ValueError(f"at least one ID must be specified for key: {key}, {self}")


@dataclass(frozen=True)
class GlobalParserActions:
    """
    Container for all :class:`.ParserAction`\\s.

    """

    actions: List[ParserAction]
    _parser_name_to_action: DefaultDict[str, List[ParserAction]] = field(
        default_factory=lambda: defaultdict(list), init=False
    )

    def __post_init__(self):
        for action in self.actions:
            for parser_name in action.parser_to_target_id_mappings.keys():
                self._parser_name_to_action[parser_name].append(action)

    def parser_behaviour(self, parser_name: str) -> Iterable[ParserAction]:
        """
        Generator that yields behaviours for a specific parser, based on the order they are
        specified in.

        :param parser_name:
        :return:
        """
        yield from self._parser_name_to_action.get(parser_name, [])

    @classmethod
    def from_json(cls, json_dict: Dict) -> "GlobalParserActions":
        return cls(
            actions=[ParserAction.from_json(x) for x in json_dict["actions"]],
        )


@dataclass(frozen=True)
class CuratedTerm:
    """
    A CuratedTerm is a means to modify the behaviour of a specific :class:`.SynonymTerm`.

    This can affect both the behaviour of :class:`kazu.ontology_preprocessing.base.OntologyParser`,
    and dictionary based NER (if using the :class:`kazu.steps.joint_ner_and_linking.explosion.ExplosionStringMatchingStep`).

    Example 1:

    The string 'ALL' is highly ambiguous. It might mean several diseases, or simply 'all'. Therefore, we want
    to add a curation as follows, so that it will only be used as a linking target and not for dictionary based NER:

    .. code-block:: python

        CuratedTerm(
            curated_synonym="ALL",
            case_sensitive=True,
            behaviour=CuratedTermBehaviour.ADD_FOR_LINKING_ONLY,
            associated_id_sets=None,
            mention_confidence=MentionConfidence.POSSIBLE,
        )


    Example 2:

    The string 'LH' is incorrectly identified as a synonym of the PLOD1 (ENSG00000083444) gene, whereas more often than not, it's actually an abbreviation of Lutenising Hormone.
    We therefore want to override the associated_id_sets to LHB (ENSG00000104826, or Lutenising Hormone Subunit Beta)

    The CuratedTerm we therefore want is:

    .. code-block:: python

        CuratedTerm(
            curated_synonym="LH",
            case_sensitive=True,
            behaviour=CuratedTermBehaviour.ADD_FOR_LINKING_ONLY,
            associated_id_sets=frozenset((EquivalentIdSet(("ENSG00000104826", "ENSEMBL")),)),
            mention_confidence=MentionConfidence.POSSIBLE,
        )

    Example 3:

    A :class:`.SynonymTerm` has an alternative synonym not referenced in the underlying ontology, and we want to add it.
    We want it to inherit all the behaviour from the original term

    .. code-block:: python

        CuratedTerm(
            curated_synonym="breast carcinoma",
            case_sensitive=True,
            source_term="breast cancer",
            behaviour=CuratedTermBehaviour.INHERIT_FROM_SOURCE_TERM,
            mention_confidence=MentionConfidence.POSSIBLE,
        )
    """

    curated_synonym: str
    mention_confidence: MentionConfidence
    behaviour: CuratedTermBehaviour
    case_sensitive: bool
    #: If specified, will override the parser defaults for the associated :class:`.SynonymTerm`\, as long as conflicts do not occur
    associated_id_sets: Optional[AssociatedIdSets] = None
    _id: bson.ObjectId = field(default_factory=bson.ObjectId, compare=False)
    #: The original term that is used as a 'seed' term for this curation.
    #: note, this is used for NER to determine how linking is performed. If you also
    #: want to use this curation as a linking target for non-dictionary based NER processes,
    #: or the term is identical to the one used in the source ontology,
    #: this should be set to None, so that a novel term_norm is calculated
    source_term: Optional[str] = None

    def __post_init__(self):
        # data validation
        if not isinstance(self.curated_synonym, str):
            raise ValueError(f"curated_synonym should be a string, {self}")

    def term_norm_for_linking(self, entity_class: str):
        norm_target = self.curated_synonym if self.source_term is None else self.source_term
        return StringNormalizer.normalize(norm_target, entity_class)

    @classmethod
    def from_json(cls, json_str: str) -> "CuratedTerm":
        json_dict = json_util.loads(json_str)
        return cls.from_dict(json_dict)

    @classmethod
    def from_dict(cls, json_dict: Dict) -> "CuratedTerm":
        if json_dict["associated_id_sets"] is None:
            frozen_assoc_id_sets = None
        else:
            assoc_id_sets = set()
            for equiv_id_set in json_dict["associated_id_sets"]:
                ids = frozenset((idx, source) for idx, source in equiv_id_set["ids_and_source"])
                equiv_id_set = EquivalentIdSet(ids)
                assoc_id_sets.add(equiv_id_set)

            frozen_assoc_id_sets = frozenset(assoc_id_sets)

        return cls(
            mention_confidence=MentionConfidence[json_dict["mention_confidence"]],
            behaviour=CuratedTermBehaviour(json_dict["behaviour"]),
            associated_id_sets=frozen_assoc_id_sets,
            case_sensitive=json_dict["case_sensitive"],
            curated_synonym=json_dict["curated_synonym"],
            source_term=json_dict["source_term"],
            _id=json_dict.get("_id", bson.ObjectId()),
        )

    def to_dict(self, preserve_structured_object_id: bool = True):
        if preserve_structured_object_id:
            oid = self._id
            as_dict = DocumentJsonUtils.obj_to_dict_repr(self)
            as_dict["_id"] = oid  # type: ignore[index,call-overload]
        else:
            as_dict = DocumentJsonUtils.obj_to_dict_repr(self)
        return as_dict

    def to_json(self) -> str:
        as_json = self.to_dict(False)
        assert isinstance(as_json, dict)
        return json.dumps(as_json)



import logging
import os
import sys
import tempfile
from copy import deepcopy
from pathlib import Path
from typing import Iterable, List

from cachetools import LFUCache
from diskcache import Cache

from kazu.data.data import Entity, SynonymTermWithMetrics
from kazu.utils.utils import get_match_entity_class_hash

logger = logging.getLogger(__name__)
kazu_model_pack_dir = os.getenv("KAZU_MODEL_PACK")
KAZU_DISK_CACHE_NAME = "kazu_disk_cache"

kazu_disk_cache: Cache
"""
We use the :class:`diskcache.Cache` concept to cache expensive to produce resources
to disk. Methods and functions can be decorated with
kazu_disk_cache.memoize()
to use this feature. Note, when used with a class method, the
default behaviour of this function is to generate a key based
on the constructor arguments of the class instance. Since these can
be large (e.g. OntologyParser), we sometimes use the ignore argument
to override this behaviour

e.g.

.. code-block:: python

    @kazu_disk_cache.memoize(ignore={0})
    def method_of_class_with_lots_of_args(self):
        ...
"""
if kazu_model_pack_dir is None:
    kazu_disk_cache_path_str = tempfile.mkdtemp(suffix=KAZU_DISK_CACHE_NAME)
    logger.warning(
        "KAZU_MODEL_PACK env variable not set. Using %s for caching. "
        "This will not be reused if the process is exited",
        kazu_disk_cache_path_str,
    )
else:
    kazu_disk_cache_path_str = str(Path(kazu_model_pack_dir).joinpath(KAZU_DISK_CACHE_NAME))
    logger.info("using disk cache at %s", kazu_disk_cache_path_str)

kazu_disk_cache = Cache(
    directory=kazu_disk_cache_path_str,
    eviction_policy="none",
    size_limit=sys.maxsize,
    cull_limit=0,
)


class EntityLinkingLookupCache:
    """
    A simple wrapper around LFUCache to reduce calls to expensive processes (e.g. bert)
    """

    def __init__(self, lookup_cache_size: int = 5000):
        self.terms_lookup_cache: LFUCache = LFUCache(lookup_cache_size)

    def update_terms_lookup_cache(self, entity: Entity, terms: Iterable[SynonymTermWithMetrics]):
        hash_val = get_match_entity_class_hash(entity)
        cache_hit = self.terms_lookup_cache.get(hash_val)
        if cache_hit is None:
            self.terms_lookup_cache[hash_val] = set(terms)

    def check_lookup_cache(self, entities: Iterable[Entity]) -> List[Entity]:
        """
        checks the cache for synonym terms. If relevant terms are found for an entity, update it
        accordingly. If not return as a list of cache misses (e.g. for further processing)

        :param entities:
        :return:
        """
        cache_misses = []
        for ent in entities:
            hash_val = get_match_entity_class_hash(ent)
            terms_from_cache = self.terms_lookup_cache.get(hash_val, set())
            if not terms_from_cache:
                cache_misses.append(ent)
            else:
                ent.update_terms(deepcopy(terms_from_cache))
        return cache_misses



from itertools import groupby
from typing import Callable, Iterable, Tuple, TypeVar, TYPE_CHECKING

# see https://github.com/python/typeshed/tree/master/stdlib/_typeshed
# In short, this is needed as
# the _typeshed package and its types do not exist at runtime
if TYPE_CHECKING:
    from _typeshed import SupportsRichComparison

Item = TypeVar("Item")
"""The type of an element in the iterable being grouped."""
# we need to use Supports RichComparison here otherwise mypy complains
# about the 'key' argument to sorted
Key = TypeVar("Key", bound="SupportsRichComparison")
"""The type of the sort key provided by the key_func.

Bound to 'SupportsRichComparison' from `_typeshed <https://github.com/python/typeshed/tree/main/stdlib/_typeshed>`_
as the keys must support comparison in order to be sorted using :func:`sorted`\\ .
"""


def sort_then_group(
    items: Iterable[Item], key_func: Callable[[Item], Key]
) -> Iterable[Tuple[Key, Iterable[Item]]]:
    sorted_items = sorted(items, key=key_func)
    yield from groupby(sorted_items, key=key_func)



from typing import List, TypedDict, Dict, NamedTuple, Iterable, Set, Tuple, DefaultDict, FrozenSet
from collections import defaultdict

import numpy as np

from kazu.data.data import Document, Entity, Section, CharSpan
from kazu.steps import Step, document_iterating_step
from sklearn.feature_extraction.text import TfidfVectorizer


class ScoredContext(NamedTuple):
    entity_class: str
    score: float
    thresh: float


class TfIdfDisambiguationEntry(NamedTuple):
    entity_class: str
    tfidf_document: np.ndarray
    tfidf_vectorizer: TfidfVectorizer
    thresh: float


class DisambiguationEntry(TypedDict):
    entity_class: str
    relevant_text: List[str]
    thresh: float


EntSectionPair = Tuple[Entity, Section]


class EntityClassTfIdfScorer:
    def __init__(self, spans_to_tfidf_disambiguator: Dict[str, List[TfIdfDisambiguationEntry]]):
        self.spans_to_tfidf_disambiguator = spans_to_tfidf_disambiguator

    @staticmethod
    def from_spans_to_sentence_disambiguator(
        spans_text_disambiguator: Dict[str, List[DisambiguationEntry]]
    ) -> "EntityClassTfIdfScorer":
        return EntityClassTfIdfScorer(
            EntityClassTfIdfScorer.build_tfidf_documents(spans_text_disambiguator)
        )

    @staticmethod
    def build_tfidf_documents(
        spans_text_disambiguator: Dict[str, List[DisambiguationEntry]]
    ) -> Dict[str, List[TfIdfDisambiguationEntry]]:
        span_to_tfidf_disambiguator = {}
        for span, disambiguation_entries in spans_text_disambiguator.items():
            span_to_tfidf_disambiguator[span] = [
                EntityClassTfIdfScorer.disambiguation_entry_to_tfidf_entry(dis_ent)
                for dis_ent in disambiguation_entries
            ]
        return span_to_tfidf_disambiguator

    @staticmethod
    def disambiguation_entry_to_tfidf_entry(
        disamb_entry: DisambiguationEntry,
    ) -> TfIdfDisambiguationEntry:
        tfidf_vectorizer = TfidfVectorizer()
        document_mat = tfidf_vectorizer.fit_transform(disamb_entry["relevant_text"])
        return TfIdfDisambiguationEntry(
            entity_class=disamb_entry["entity_class"],
            tfidf_document=document_mat.data,
            tfidf_vectorizer=tfidf_vectorizer,
            thresh=disamb_entry["thresh"],
        )

    @staticmethod
    def tfidf_score(
        ent_context: str, tfidf_disambig_entry: TfIdfDisambiguationEntry
    ) -> ScoredContext:
        vectorizer = tfidf_disambig_entry.tfidf_vectorizer
        doc_mat = tfidf_disambig_entry.tfidf_document
        vector_context = vectorizer.transform([ent_context])
        score = np.squeeze(np.asarray(vector_context.dot(doc_mat.T))).item()
        return ScoredContext(
            entity_class=tfidf_disambig_entry.entity_class,
            score=score,
            thresh=tfidf_disambig_entry.thresh,
        )

    def score_entity_context(self, ent_span: str, ent_context: str) -> Iterable[ScoredContext]:
        """
        Score the entity context against the TfIdf documents specified for the entity's span

        :param ent_span:
        :param ent_context:
        :return:
        """
        if ent_span not in self.spans_to_tfidf_disambiguator:
            yield from ()
        else:
            tfidf_disambiguation_entries = self.spans_to_tfidf_disambiguator[ent_span]
            for tfidf_disambiguation_entry in tfidf_disambiguation_entries:
                scored_context = self.tfidf_score(ent_context, tfidf_disambiguation_entry)
                yield scored_context


class EntityClassDisambiguationStep(Step):
    def __init__(self, context: Dict[str, List[DisambiguationEntry]]):
        """
        Optionally disambiguates the entity class (anatomy, drug, etc.) of entities that exactly share a span in a
        document.

        For example, "UCB" could refer to "umbilical cord blood" an anatomical entity, or the pharmaceutical company
        UCB, a corporate entity. An expected context might be "umbilical pregnancy blood baby placenta..." in the former
        case, or "company business..." in the latter. Multiple expected contexts (disambiguation entries) should be
        provided to allow this step to choose the best matching entity class for an entity span. A tf-idf model is built
        to correlate an entity's actual textual context with the provided expected context, and provided thresholds are
        used to allow the tf-idf model to choose the most suitable entity class.

        :param context: Specifies terms to disambiguate along with an expected textual context around those terms.
        """
        self.spans_to_disambiguate = set(context.keys())
        self.entity_class_scorer = EntityClassTfIdfScorer.from_spans_to_sentence_disambiguator(
            context
        )

    @staticmethod
    def sentence_context_for_entity(entity: Entity, section: Section, window: int = 3) -> str:
        sent_spans = list(section.sentence_spans)
        if len(sent_spans) == 0:
            return section.text
        else:
            idx = next(
                idx
                for idx, sent_span in enumerate(sent_spans)
                if any(
                    entity_span.is_completely_overlapped(sent_span) for entity_span in entity.spans
                )
            )
            context_start_idx = max(0, idx - int(window / 2))
            context_end_idx = min(len(sent_spans), idx + int(window / 2)) + 1
            sentence_context_spans = sent_spans[context_start_idx:context_end_idx]
            return section.text[sentence_context_spans[0].start : sentence_context_spans[-1].end]

    def spangrouped_ent_section_pairs(self, doc: Document) -> Iterable[List[EntSectionPair]]:
        spans_to_ents_and_sections: DefaultDict[
            Tuple[int, FrozenSet[CharSpan]], List[EntSectionPair]
        ] = defaultdict(list)
        for section_idx, section in enumerate(doc.sections):
            for ent in section.entities:
                if ent.match in self.spans_to_disambiguate:
                    spans_to_ents_and_sections[(section_idx, ent.spans)].append((ent, section))
        return spans_to_ents_and_sections.values()

    @document_iterating_step
    def __call__(self, doc: Document) -> None:
        drop_set: Dict[Section, Set[Entity]] = defaultdict(set)
        for spansharing_ent_section_pairs in self.spangrouped_ent_section_pairs(doc):
            if len(spansharing_ent_section_pairs) == 1:
                # only a single entity; span(s) are unambiguous for entity class
                continue

            representative_ent, representative_section = spansharing_ent_section_pairs[0]
            ents: List[Entity] = [ent for ent, section in spansharing_ent_section_pairs]
            class_to_ents = defaultdict(set)
            for ent in ents:
                class_to_ents[ent.entity_class].add(ent)

            ent_match = representative_ent.match
            ent_sentence_context_str = self.sentence_context_for_entity(
                representative_ent, representative_section
            )
            scored_contexts = self.entity_class_scorer.score_entity_context(
                ent_span=ent_match, ent_context=ent_sentence_context_str
            )

            acceptable_scores = (
                scored_context
                for scored_context in scored_contexts
                if scored_context.score >= scored_context.thresh
            )
            best_score = max(
                acceptable_scores, default=None, key=lambda scored_context: scored_context.score
            )
            if best_score is not None:
                best_match_ents = class_to_ents[best_score.entity_class]
                drop_set[representative_section].update(
                    (ent for ent in ents if ent not in best_match_ents)
                )
            else:
                drop_set[representative_section].update(ents)

        for section, drop_entities in drop_set.items():
            section.entities = [ent for ent in section.entities if ent not in drop_entities]
